/* Copyright (c) 2010-2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.main;

import java.awt.Color;
import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.util.Collection;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;

import se.llbit.chunky.renderer.RenderManager;
import se.llbit.chunky.renderer.ui.Chunk3DView;
import se.llbit.chunky.resources.TexturePackLoader;
import se.llbit.chunky.ui.ChunkMap;
import se.llbit.chunky.ui.ChunkyLogHandler;
import se.llbit.chunky.ui.ChunkyFrame;
import se.llbit.chunky.ui.Controls;
import se.llbit.chunky.world.Block;
import se.llbit.chunky.world.Chunk;
import se.llbit.chunky.world.Chunk.Renderer;
import se.llbit.chunky.world.listeners.ChunkDiscoveryListener;
import se.llbit.chunky.world.listeners.RegionDiscoveryListener;
import se.llbit.chunky.world.ChunkParser;
import se.llbit.chunky.world.ChunkPosition;
import se.llbit.chunky.world.ChunkSelectionTracker;
import se.llbit.chunky.world.ChunkView;
import se.llbit.chunky.world.DeleteChunksJob;
import se.llbit.chunky.world.EmptyChunkView;
import se.llbit.chunky.world.EmptyWorld;
import se.llbit.chunky.world.Region;
import se.llbit.chunky.world.RegionParser;
import se.llbit.chunky.world.World;
import se.llbit.chunky.world.WorldRenderer;
import se.llbit.chunky.world.WorldScanner;
import se.llbit.util.ProgramProperties;

/**
 * The entry point for the Chunky application.
 * 
 * @author Jesper Öqvist (jesper@llbit.se)
 */
public class Chunky implements ChunkDiscoveryListener,
		RegionDiscoveryListener {
	
	public static final double ORIGIN_X = 0;
	public static final double ORIGIN_Z = 0;
	public static final int BLOCK_SCALE_MIN = 0;
	public static final int BLOCK_SCALE_MAX = 32;
	public static final int DEFAULT_BLOCK_SCALE = 4;
	
	private int chunkScale = DEFAULT_BLOCK_SCALE*16;
	
	private World world = new EmptyWorld();
	
	private ChunkParser chunkParser = new ChunkParser();
	private RegionParser regionParser = new RegionParser();
	private WorldScanner worldScanner;

	private int currentDimension = 0;
	private Chunk.Renderer chunkRenderer = Chunk.surfaceRenderer;
	private WorldRenderer worldRenderer = new WorldRenderer();
	protected ChunkSelectionTracker chunkSelection = new ChunkSelectionTracker();

	protected boolean ctrlModifier = false;
	protected boolean shiftModifier = false;

	private Chunk3DView chunk3DView = null;
	private ChunkyFrame frame;

	private ChunkView map = EmptyChunkView.instance;
	private ChunkView minimap = EmptyChunkView.instance;
	
	private int mapWidth = ChunkMap.DEFAULT_WIDTH;
	private int mapHeight = ChunkMap.DEFAULT_HEIGHT;
	
	private int minimapWidth = Minimap.DEFAULT_WIDTH;
	private int minimapHeight = Minimap.DEFAULT_HEIGHT;
	
	public static Logger logger;

	public static final String getAppName() {
		return String.format("%s %s (%s)", //$NON-NLS-1$
			Messages.getString("Chunky.appname"), //$NON-NLS-1$
			Messages.getString("Chunky.version"), //$NON-NLS-1$
			Messages.getString("Chunky.revision")); //$NON-NLS-1$
	}
	
	static {
		// setup main logger
		logger = Logger.getLogger("chunky");
	}
	
	/**
	 * Create a new instance of the application GUI.
	 * @param args Command-line arguments
	 */
	@SuppressWarnings("unused")
	private Chunky(String[] args) {
		
		boolean selectedWorld = false;
		String sceneName = null;
		String texturePack = null;
		int numThreads = 6;
		File worldDir = null;
		for (int i = 0; i < args.length; ++i) {
			if (args[i].equals("-texture") && args.length > i+1) {
				i += 1;
				texturePack = args[i];
			} else if (args[i].equals("-watermark")) {
				RenderManager.useWatermark = true;
			} else if (args[i].equals("-render")) {
				i += 1;
				sceneName = args[i];
			} else if (args[i].equals("-threads")) {
				i += 1;
				numThreads = Math.max(1, Integer.parseInt(args[i]));
				numThreads = Math.min(1, 20);
			} else if (!selectedWorld) {
				worldDir = new File(args[i]);
			}
		}
		
		if (texturePack != null) {
			TexturePackLoader.loadTexturePack(new File(texturePack));
		} else {
			String lastTexturePack = ProgramProperties.getProperty("lastTexturePack");
			if (lastTexturePack != null)
				TexturePackLoader.loadTexturePack(new File(lastTexturePack));
			else
				TexturePackLoader.loadTexturePack(getMinecraftJar());
		}
		
		// TODO
		/*if (sceneName != null) {
			chunk3DView = new Chunk3DView(Chunky.this,
					new LinkedList<ChunkPosition>(), numThreads, 0);
			try {
				chunk3DView.getRenderController().loadScene(null, new File("Scenes/" + sceneName + ".cvf"));
			} catch (IOException e) {
				e.printStackTrace();
			} catch (SceneLoadingError e) {
				e.printStackTrace();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			return;
		}*/
		
		// load the world
		if (worldDir != null && World.isWorldDir(worldDir)) {
			loadWorld(new World(worldDir));
		} else {
			String lastWorld = ProgramProperties.getProperty("lastWorld");
			File lastWorldDir = lastWorld != null ? new File(lastWorld) : null;
			if (lastWorldDir != null && World.isWorldDir(lastWorldDir)) {
				loadWorld(new World(lastWorldDir));
			}
		}
		
		// Start the worker threads
		chunkParser.start();
		regionParser.start();
		
		// Create UI in the event dispatch thread
		try {
			SwingUtilities.invokeAndWait(new Runnable() {
				@Override
				public void run() {
					buildUI();
				}
			});
		} catch (InterruptedException e) {
			logger.log(Level.SEVERE, "Failed to initialize UI", e);
		} catch (InvocationTargetException e) {
			logger.log(Level.SEVERE, "Failed to initialize UI", e);
		}
	}
	
	protected void buildUI() {
		ChunkyLogHandler handler = new ChunkyLogHandler();
		logger.addHandler(handler);
		
		frame = new ChunkyFrame(this);
		frame.initComponents();
		frame.setVisible(true);
	}
	
	public synchronized void reloadWorld() {
		if (worldScanner != null)
			worldScanner.interrupt();
		
		chunkSelection.clearSelection();
		
		world.reload();
		worldScanner = new WorldScanner(world);
		worldScanner.start();
		
		if (frame != null) {
			viewUpdated();
			frame.getMap().redraw();
			frame.getMinimap().redraw();
		}
	}

	public synchronized void loadWorld(World newWorld) {
		
		regionParser.clearQueue();
		chunkParser.clearQueue();
		if (worldScanner != null)
			worldScanner.interrupt();
		
		chunkSelection.clearSelection();
		
		// dispose old world
		world.dispose();
		
		world = newWorld;
		world.addChunkDeletionListener(chunkSelection);
		world.addChunkDiscoveryListener(this);
		world.addRegionDiscoveryListener(this);

		// dimension must be set before chunks are loaded
		world.setDimension(currentDimension);
		
		if (world.havePlayerPos()) {
			setView(
					world.playerPosX() / 16,
					world.playerPosZ() / 16);
		} else {
			setView(Chunky.ORIGIN_X, Chunky.ORIGIN_Z);
		}
		
		worldScanner = new WorldScanner(world);
		worldScanner.start();
		
		ProgramProperties.setProperty("lastWorld",
				world.getWorldDirectory().getAbsolutePath());
		
		if (frame != null) {
			frame.worldLoaded(world);
			viewUpdated();
			frame.getMap().redraw();
			frame.getMinimap().redraw();
		}
	}
	
	/**
	 * Called when the map view has changed.
	 */
	public synchronized void viewUpdated() {
		minimap = new ChunkView(map.x, map.z, minimapWidth, minimapHeight, 1);
		
		for (int rx = Math.min(minimap.rx0, map.prx0);
				rx <= Math.max(minimap.rx1, map.prx1); ++rx) {
			
			for (int rz = Math.min(minimap.rz0, map.prz0);
					rz <= Math.max(minimap.rz1, map.prz1); ++rz) {
				
				Region region = world.getRegion(ChunkPosition.get(rx, rz));
				if (!region.isEmpty()
						&& (map.isVisible(region) || minimap.isVisible(region))) {
					if (!region.isParsed()) {
						regionParser.addRegion(region);
					} else {
						// parse visible chunks
						region.preloadChunks(map, chunkParser);
					}
				}
			}
		}
		
		if (frame != null) {
			frame.getMap().viewUpdated(map);
			frame.getMinimap().viewUpdated(minimap);
		}
	}

	public static void main(final String[] args) {
		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception e) {
			logger.warning("Failed to set native Look and Feel");
		}
		
		// Create UI on the event dispatch thread
		Chunky chunky = new Chunky(args);
		chunky.refreshLoop();
	}
	
	/**
	 * Periodically check for changes in the world state and
	 * refresh the view if necessary.
	 */
	private void refreshLoop() {
		try {
			while (!Thread.interrupted()) {
				frame.refresh();
				Thread.sleep(125);
			}
		} catch (InterruptedException e) {
		}
	}

	public synchronized void setRenderer(Chunk.Renderer renderer) {
		this.chunkRenderer = renderer;
		getMap().repaint();
	}

	public void open3DView() {
		if (chunk3DView == null || !chunk3DView.isDisplayable())
			chunk3DView = new Chunk3DView(Chunky.this, chunkSelection.getSelection(),
					getControls().getNumThreads(),
					getControls().getWaterWorld() ?
							getControls().getWaterHeight() : 0);
	}

	public synchronized void selectChunk(int cx, int cz) {
		chunkSelection.selectChunk(world, cx, cz);
		getControls().setChunksSelected(chunkSelection.numSelectedChunks() > 0);
	}
	
	public synchronized void setView(final double cx, final double cz) {
		map = new ChunkView(cx, cz, mapWidth, mapHeight, chunkScale);
		if (frame != null) {
			getControls().setPosition(cx, getLayer(), cz);
		}
		viewUpdated();
	}

	public synchronized void setLayer(int value) {
		int layerNew = Math.max(0, Math.min(Chunk.Y_MAX-1, value));
		if (layerNew != world.currentLayer()) {
			chunkParser.clearQueue();
			world.setCurrentLayer(layerNew);
			if (chunkRenderer == Chunk.layerRenderer)
				getMap().redraw();
			
			// force the chunks to redraw
			viewUpdated();
		}
		getControls().setLayer(world.currentLayer());
	}
	
	public synchronized int getLayer() {
		return world.currentLayer();
	}

	public void deleteSelectedChunks() {
		Object[] options = {Messages.getString("Chunky.Cancel_lbl"), //$NON-NLS-1$
				Messages.getString("Chunky.AcceptDelete_lbl")}; //$NON-NLS-1$
		int n = JOptionPane.showOptionDialog(null,
				Messages.getString("Chunky.DeleteDialog_msg"), //$NON-NLS-1$
				Messages.getString("Chunky.DeleteDialog_title"), //$NON-NLS-1$
				JOptionPane.YES_NO_OPTION,
				JOptionPane.WARNING_MESSAGE,
				null,
				options,
				options[0]);
		if (n == 1) {

			Collection<ChunkPosition> selected = chunkSelection.getSelection();
			ProgressPanel progress = getControls().getProgressPanel();
			if (!selected.isEmpty() && !progress.isBusy()) {
				DeleteChunksJob job = new DeleteChunksJob(world, selected, progress);
				job.start();
			}
		}
	}

	public void setDimension(int value) {
		if (value != currentDimension) {
			currentDimension = value;
			loadWorld(world);
		}
	}

	public synchronized void clearSelectedChunks() {
		chunkSelection.clearSelection();
		getControls().setChunksSelected(chunkSelection.numSelectedChunks() > 0);
	}
	
	public static File getMinecraftDirectory() {
		String home = System.getProperty("user.home", ".");   //$NON-NLS-1$ //$NON-NLS-2$
		String os = System.getProperty("os.name").toLowerCase();  //$NON-NLS-1$
		
		if (os.contains("win")) {  //$NON-NLS-1$
			String appdata = System.getenv("APPDATA");	//$NON-NLS-1$
			if (appdata != null)
				return new File(appdata, ".minecraft");	//$NON-NLS-1$
			else
				return new File(home, ".minecraft");  //$NON-NLS-1$
		} else if (os.contains("mac")) {  //$NON-NLS-1$
			return new File(home, "Library/Application Support/minecraft");	//$NON-NLS-1$
		} else {
			return new File(home, ".minecraft");  //$NON-NLS-1$
		}
	}

	public static File getSavesDirectory() {
		return new File(getMinecraftDirectory(), "saves");
	}
	
	public static File getTexturePacksDirectory() {
		return new File(getMinecraftDirectory(), "texturepacks");
	}
	
	public Color getHighlightColor() {
		return worldRenderer.getHighlightColor();
	}
	
	public Block getHighlightBlock() {
		return worldRenderer.getHighlightBlock();
	}

	public void setHighlightEnable(boolean value) {
		worldRenderer.setHighlightEnabled(value);
		getMap().redraw();
	}
	
	public boolean isHighlightEnabled() {
		return worldRenderer.isHighlightEnabled();
	}

	public void highlightBlock(Block hlBlock) {
		worldRenderer.highlightBlock(hlBlock);
		getMap().redraw();
	}

	public void setHighlightColor(Color newColor) {
		worldRenderer.setHighlightColor(newColor);
		getMap().redraw();
	}

	public File getWorldDirectory() {
		return world.getWorldDirectory();
	}

	public synchronized void exportZip(File targetFile, ProgressPanel progress) {
		if (!progress.isBusy()) {
			if (targetFile.exists()) {
				Object[] options = {Messages.getString("Chunky.Cancel_lbl"), //$NON-NLS-1$
						Messages.getString("Chunky.AcceptOverwrite_lbl")}; //$NON-NLS-1$
				int n = JOptionPane.showOptionDialog(null,
						String.format(Messages.getString("Chunky.Confirm_overwrite_msg"), //$NON-NLS-1$
								targetFile.getName()),
						Messages.getString("Chunky.Confirm_overwrite_title"), //$NON-NLS-1$
						JOptionPane.YES_NO_OPTION,
						JOptionPane.WARNING_MESSAGE,
						null,
						options,
						options[0]);
				if (n != 1)
					return;
			}
			new ZipExportJob(world, chunkSelection.getSelection(), targetFile, progress).start();
		}
	}

	public void renderView(File targetFile, ProgressPanel progress) {
		if (!progress.isBusy()) {
			if (targetFile.exists()) {
				Object[] options = {Messages.getString("Chunky.Cancel_lbl"), //$NON-NLS-1$
						Messages.getString("Chunky.AcceptOverwrite_lbl")}; //$NON-NLS-1$
				int n = JOptionPane.showOptionDialog(null,
						String.format(Messages.getString("Chunky.Confirm_overwrite_msg"), //$NON-NLS-1$
								targetFile.getName()),
						Messages.getString("Chunky.Confirm_overwrite_title"), //$NON-NLS-1$
						JOptionPane.YES_NO_OPTION,
						JOptionPane.WARNING_MESSAGE,
						null,
						options,
						options[0]);
				if (n != 1)
					return;
			}
			
			if (progress.tryStartJob()) {
				progress.setJobName("PNG export");
				progress.setJobSize(1);
				getMap().renderPng(targetFile);
				progress.finishJob();
			}
		}
	}
	
	public void renderDimension(File targetFile, ProgressPanel progress) {
		if (!progress.isBusy()) {
			if (targetFile.exists()) {
				Object[] options = {Messages.getString("Chunky.Cancel_lbl"), //$NON-NLS-1$
						Messages.getString("Chunky.AcceptOverwrite_lbl")}; //$NON-NLS-1$
				int n = JOptionPane.showOptionDialog(null,
						String.format(Messages.getString("Chunky.Confirm_overwrite_msg"), //$NON-NLS-1$
								targetFile.getName()),
						Messages.getString("Chunky.Confirm_overwrite_title"), //$NON-NLS-1$
						JOptionPane.YES_NO_OPTION,
						JOptionPane.WARNING_MESSAGE,
						null,
						options,
						options[0]);
				if (n != 1)
					return;
			}
			
			if (world.numRegions() != 0) {
				// don't render PNG if world is empty
				new RenderDimensionJob(world, targetFile, progress).start();
			} else {
				Chunky.logger.warning("Can not render dimension: no chunks to render!");
			}
		}
	}

	public World getWorld() {
		return world;
	}

	public Collection<ChunkPosition> getSelectedChunks() {
		return chunkSelection.getSelection();
	}

	public boolean getShiftModifier() {
		return shiftModifier;
	}

	public boolean getCtrlModifier() {
		return ctrlModifier;
	}

	public void selectChunks(int cx0, int cx1, int cz0, int cz1) {
		if (!ctrlModifier)
			chunkSelection.selectChunks(world, cx0, cz0, cx1, cz1);
		else
			chunkSelection.deselectChunks(world, cx0, cz0, cx1, cz1);
		
		getControls().setChunksSelected(chunkSelection.numSelectedChunks() > 0);
	}

	public Controls getControls() {
		return frame.getControls();
	}

	public ChunkSelectionTracker getChunkSelection() {
		return chunkSelection;
	}

	public WorldRenderer getWorldRenderer() {
		return worldRenderer;
	}

	public void setCtrlModifier(boolean value) {
		ctrlModifier = value;
	}

	public void setShiftModifier(boolean value) {
		shiftModifier = value;
	}

	public ChunkyFrame getFrame() {
		return frame;
	}

	public ChunkMap getMap() {
		return frame.getMap();
	}

	public Renderer getChunkRenderer() {
		return chunkRenderer;
	}

	public Minimap getMinimap() {
		return frame.getMinimap();
	}
	
	@Override
	public synchronized void regionDiscovered(Region region) {
		if (map.shouldPreload(region) || minimap.isVisible(region)) {
			regionParser.addRegion(region);
		}
	}

	@Override
	public void chunksDiscovered(Collection<Chunk> chunks) {
		for (Chunk chunk: chunks) {
			if (map.shouldPreload(chunk))
				chunkParser.addChunk(chunk);
		}
	}

	public boolean isLoading() {
		return chunkParser.isWorking() || regionParser.isWorking();
	}
	
	public synchronized void setScale(int value) {
		int scaleNew = Math.max(BLOCK_SCALE_MIN,
				Math.min(BLOCK_SCALE_MAX, value));
		scaleNew = scaleNew == 0 ? 1 : scaleNew*16;
		if (scaleNew != chunkScale) {
			chunkScale = scaleNew;
			getControls().setScale(getScale());
			setView(map.x, map.z);
		}
	}

	public int getScale() {
		return chunkScale/16;
	}

	public void viewDragged(int dx, int dy) {
		moveView(dx / (double) chunkScale,
				dy / (double) chunkScale);
	}
	
	/**
	 * Move the map view
	 * @param dx
	 * @param dz
	 */
	public synchronized void moveView(double dx, double dz) {
		setView(map.x + dx, map.z + dz);
	}

	public ChunkView getMapView() {
		return map;
	}

	public void mapResized(int width, int height) {
		if (width != mapWidth || height != mapHeight) {
			mapWidth = width;
			mapHeight = height;
			setView(map.x, map.z);
		}
	}

	public synchronized void onMouseWheelMotion(int diff) {
		if (!ctrlModifier) {
			setLayer(getLayer() + diff);
		} else {
			setScale(getScale() - diff);
		}
	}

	public ChunkView getMinimapView() {
		return minimap;
	}

	public void minimapResized(int width, int height) {
		if (width != minimapWidth || height != minimapHeight) {
			minimapWidth = width;
			minimapHeight = height;
			viewUpdated();
		}
	}

	public static final File getMinecraftJar() {
		return new File(Chunky.getMinecraftDirectory(), "bin/minecraft.jar");
	}
}
