/* Copyright (c) 2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.renderer;

import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.FileDialog;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferInt;
import java.awt.image.Raster;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Collection;
import java.util.LinkedList;
import java.util.Random;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.logging.Level;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import javax.imageio.ImageIO;
import javax.swing.JOptionPane;

import se.llbit.chunky.main.CenteredFileDialog;
import se.llbit.chunky.main.Chunky;
import se.llbit.chunky.main.SceneLoadingError;
import se.llbit.chunky.main.Messages;
import se.llbit.chunky.renderer.RenderWorker;
import se.llbit.chunky.renderer.ui.Canvas3D;
import se.llbit.chunky.renderer.ui.Chunk3DView;
import se.llbit.chunky.renderer.ui.RenderControls;
import se.llbit.chunky.world.ChunkPosition;
import se.llbit.chunky.world.World;
import se.llbit.math.Color;
import se.llbit.math.Ray;
import se.llbit.math.Vector3d;
import se.llbit.math.Ray.RayPool;
import se.llbit.resources.ImageLoader;
import se.llbit.util.ProgramProperties;
import se.llbit.util.VectorPool;

public class RenderManager extends Thread {
	public static final int NUM_RENDER_THREADS_MIN = 1;
	public static final int NUM_RENDER_THREADS_MAX = 20;
	public static final int NUM_RENDER_THREADS_DEFAULT = Runtime.getRuntime().availableProcessors();
	
	public static final int MIN_CANVAS_WIDTH = 20;
	public static final int MIN_CANVAS_HEIGHT = 20;

	public static final int SEG_X = 6;
	public static final int SEG_Y = 6;
	private static final int NUM_JOBS = SEG_X * SEG_Y;
	private static final int SPP = 1;
	private final int numThreads;
	public int dumpInterval = 500;
	public boolean saveDumps = true;
	private Canvas3D canvas;
	private Thread[] workers;
	private BufferedImage buffer;
	private double[][][][] samples = new double[NUM_JOBS][][][];
	private int numSamples = 0;
	int canvasWidth;
	int canvasHeight;
	private int[] width = new int[NUM_JOBS];
	private int[] height = new int[NUM_JOBS];
	private int[] xoffset = new int[NUM_JOBS];
	private int[] yoffset = new int[NUM_JOBS];
	
	Collection<RenderStateListener> listeners = new LinkedList<RenderStateListener>();

	/**
 	 * The modifiable scene.
 	 */
	private Scene scene;

	/**
 	 * The buffered scene is only updated between
 	 * render jobs.
 	 */
	private Scene bufferedScene;
	
	/**
	 * Bitmask for worker threads.
	 */
	private long tickets = 0;
	
	/**
	 * Bitmask for done tiles.
	 */
	private long done;
	
	private RayPool[] rayPool;
	private VectorPool[] vectorPool;
	private Random[] random;
	
	private Chunk3DView view;
	private boolean isReady = false;
	
	private long frameTime;
	private boolean statusTextUpdated;
	private String statusText;
	
	public final int waterHeight;

	private static BufferedImage watermark = ImageLoader.get("watermark.png");
	public static boolean useWatermark = false;
	
	private static final Font infoFont = new Font("Sans serif", Font.BOLD, 11);
	private static FontMetrics fontMetrics;
	
	ByteArrayOutputStream statusBuffer = new ByteArrayOutputStream() {
		@Override
		public void flush() throws IOException {
			super.flush();
			
			synchronized (RenderManager.this) {
				statusText = toString();
				statusTextUpdated = true;
				RenderManager.this.notifyAll();
			}
		}
	};

	private PrintStream status = new PrintStream(statusBuffer);
	
	private final Lock sceneLock = new ReentrantLock();
	
	public RenderManager(World world, Chunk3DView view,
			Canvas3D canvas, Collection<ChunkPosition> selectedChunks,
			int nRenderThreads, int waterHeight) {
		super("Render Controller");
		
		this.waterHeight = waterHeight;
		this.view = view;
		this.numThreads = nRenderThreads;
		this.canvas = canvas;
		
		scene = new Scene(this, world);
		bufferedScene = new Scene(scene);
		
		canvasWidth = canvas.getWidth();
		canvasHeight = canvas.getHeight();
		initBuffers();
		
		// init worker data pools
		rayPool = new RayPool[numThreads];
		vectorPool = new VectorPool[numThreads+1];
		random = new Random[numThreads+1];
		
		long seed = System.currentTimeMillis();
		for (int i = 0; i < numThreads; ++i) {
			rayPool[i] = new Ray.RayPool();
			vectorPool[i] = new VectorPool();
			random[i] = new Random(seed + i);
		}
		vectorPool[numThreads] = new VectorPool();
		random[numThreads] = new Random(seed + numThreads);
		
		updateCanvas("Loading chunks...");
		scene.loadChunks(selectedChunks);
		scene.centerCamera();

		// start worker threads
		workers = new Thread[numThreads];
		for (int i = 0; i < numThreads; ++i) {
			workers[i] = new RenderWorker(this, i);
			workers[i].start();
		}
	}

	public void run() {
		long totalSamples = 0;
		int prevNumSamples;
		try {
			try {
				bufferedScene.set(scene);

				long frameStart = System.currentTimeMillis();
				giveTickets();
				
				synchronized (this) {
					isReady = true;
					notifyAll();
				}

				while (!isInterrupted()) {
					
					boolean pathTrace = bufferedScene.pathTrace();
					
					if (numSamples == 0) {
						frameTime = 0;

						// enable JIT for the first frame
						java.lang.Compiler.enable();
					}
					
					waitOnWorkers();
					bufferedScene.previewCount -= 1;// TODO
					frameTime += System.currentTimeMillis() - frameStart;

					// disable JIT for subsequent frames
					java.lang.Compiler.disable();

					numSamples += pathTrace ? SPP : 1;
					prevNumSamples = numSamples;

					if (pathTrace) {
						if (scene.shouldRefresh()) {
							fireRenderStateEvent(scene.pathTrace(), scene.isPaused());
							bufferedScene.set(scene);
							numSamples = 0;
						}
						if (!scene.isPaused()) {
							frameStart = System.currentTimeMillis();
							giveTickets();
							updateCanvas();
						} else {
							updateCanvas();
							scene.pauseWait();
							frameStart = System.currentTimeMillis();
							giveTickets();
						}
					} else {
						
						numSamples = 0;
						updateCanvas();
						if (bufferedScene.previewCount <= 0) {
							scene.waitOnRefreshRequest();
							fireRenderStateEvent(scene.pathTrace(), scene.isPaused());
							bufferedScene.set(scene);
						} else if (scene.shouldRefresh()) {
							fireRenderStateEvent(scene.pathTrace(), scene.isPaused());
							bufferedScene.set(scene);
						}
						frameStart = System.currentTimeMillis();
						giveTickets();
					}
					
					totalSamples = prevNumSamples
						* ((long) (canvasWidth * canvasHeight));
					
					int seconds = (int) ((frameTime / 1000) % 60);
					int minutes = (int) ((frameTime / 60000) % 60);
					int hours = (int) (frameTime / 3600000);
					
					StringBuffer buf = new StringBuffer();
					buf.append(String.format("canvas size: %dx%d%n", canvasWidth, canvasHeight));
					buf.append(String.format("rendered for %d hours, %d minutes, %d seconds%n",
							hours, minutes, seconds));
					buf.append(String.format("%d samples per second%n",
							(int) (totalSamples / (frameTime / 1000.0))));
					buf.append(String.format("%dk total samples%n",
							totalSamples / 1000));
					buf.append(String.format("%d samples per pixel%n", prevNumSamples));
					
					statusBuffer.reset();
					status.print(buf);
					status.flush();
					
					if (saveDumps && (prevNumSamples % dumpInterval == 0)) {
						frameTime += System.currentTimeMillis() - frameStart;
						try {
							saveFrame(prevNumSamples);

							backupFile(scene.name() + ".txt");
							PrintStream out = new PrintStream(
									new FileOutputStream(
											new File(createScenesDirectory(),
											scene.name() + ".txt")));
							out.print(buf.toString());
							out.close();

							// create dump
							backupFile(scene.name() + ".dump");
							backupFile(scene.name() + ".cvf");
							saveScene(new File(createScenesDirectory(),
									scene.name() + ".cvf"));
						} catch (IOException e) {
							e.printStackTrace();
						}
						frameStart = System.currentTimeMillis();
					}
				}
			} catch (InterruptedException e) {
			}
		} catch (Throwable e) {
			Chunky.logger.log(Level.SEVERE,
					"Uncaught exception in render controller", e);
		}
		
		for (int i = 0; i < numThreads; ++i) {
			workers[i].interrupt();
		}
	}

	private void backupFile(String fileName) {
		File renderDir = createScenesDirectory();
		File file = new File(renderDir, fileName);
		if (file.exists()) {
			// try to create backup
			// it's not a biggie if we can't
			File backup = new File(renderDir, fileName  + ".backup");
			if (backup.exists())
				backup.delete();
			file.renameTo(new File(renderDir, fileName + ".backup"));
		}
	}
	
	public static File createScenesDirectory() {
		File scenesDir = new File("Scenes");
		while (!scenesDir.exists() || !scenesDir.isDirectory()) {
			if (!scenesDir.mkdir()) {
				JOptionPane.showMessageDialog(null,
						"<html>Could not create or open the directory "+scenesDir.getAbsolutePath() +
						"<br>This directory is used by Chunky to save render dumps and scene descriptions in." +
						"<br>Please ensure that this directory exists and is writable!",
						"Could not open Renders directory",
						JOptionPane.ERROR_MESSAGE);
			} else {
				break;
			}
		}
		return scenesDir;
	}

	private void saveFrame(int sampleCount) throws IOException {
		String fileName = scene.getSceneName() + "-" + sampleCount + ".png";
		status.println("Saving frame " + fileName);
		status.flush();
		addWatermark();
		ImageIO.write(buffer, "png", new File(fileName));
		status.println("Frame saved");
		status.flush();
	}

	/**
	 * Add a watermark to the image buffer.
	 */
	private void addWatermark() {
		if (useWatermark) {
			Graphics g = buffer.getGraphics();
			g.drawImage(watermark, buffer.getWidth() - watermark.getWidth(),
					buffer.getHeight() - watermark.getHeight(), null);
			g.dispose();
		}
	}

	void saveDump(String fileName) {
		DataOutputStream out = null;
		try {
			status.println("Saving dump " + fileName);
			status.flush();
			out = new DataOutputStream(
					new GZIPOutputStream(new FileOutputStream(
					new File(createScenesDirectory(), fileName))));
			out.writeInt(canvasWidth);
			out.writeInt(canvasHeight);
			out.writeInt(numSamples);
			out.writeLong(frameTime);
			int jobW = canvasWidth / SEG_X;
			int jobH = canvasHeight / SEG_Y;
			for (int y = 0; y < canvasHeight; ++y) {
				int jobOffset = SEG_X * Math.min(SEG_Y-1, y/jobH);
				for (int x = 0; x < canvasWidth; ++x) {
					int jobId = jobOffset + Math.min(SEG_X-1, x/jobW);
					int xo = xoffset[jobId];
					int yo = yoffset[jobId];
					for (int i = 0; i < 3; ++i) {
						out.writeDouble(samples[jobId][x-xo][y-yo][i]);
					}
				}
			}
			status.println("Dump saved");
			status.flush();
		} catch (IOException e) {
			status.println("IO exception while saving dump!");
			status.flush();
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (IOException e) {
				}
			}
		}
	}
	
	void loadDump(String fileName) {
		DataInputStream in = null;
		try {
			status.println("Loading dump " + fileName);
			status.flush();
			in = new DataInputStream(
					new GZIPInputStream(new FileInputStream(
					new File(createScenesDirectory(), fileName))));
			int dumpWidth = in.readInt();
			int dumpHeight= in.readInt();
			if (dumpWidth != canvasWidth || dumpHeight != canvasHeight) {
				status.println("Dump discarded! Incorrect widht/height.");
				status.flush();
				in.close();
				return;
			}
			numSamples = in.readInt();
			frameTime = in.readLong();
			int jobW = canvasWidth / SEG_X;
			int jobH = canvasHeight / SEG_Y;
			for (int y = 0; y < canvasHeight; ++y) {
				int jobOffset = SEG_X * Math.min(SEG_Y-1, y/jobH);
				for (int x = 0; x < canvasWidth; ++x) {
					int jobId = jobOffset + Math.min(SEG_X-1, x/jobW);
					int xo = xoffset[jobId];
					int yo = yoffset[jobId];
					for (int i = 0; i < 3; ++i) {
						samples[jobId][x-xo][y-yo][i] = in.readDouble();
					}
				}
			}
			status.println("Dump loaded");
			status.flush();
			bufferedScene.set(scene);
			for (int jobId = 0; jobId < NUM_JOBS; ++jobId) {
				finalizeJob(jobId);
			}
		} catch (IOException e) {
			status.println("Could not load dump: Missing file or incorrect format!");
			status.flush();
		} finally {
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
				}
			}
		}
		
		updateCanvas("Loading chunks...");
	}

	private void initBuffers() {
		int h = canvasHeight/SEG_Y;
		int w = canvasWidth/SEG_X;
		
		for (int y = 0; y < SEG_Y; ++y) {
			for (int x = 0; x < SEG_X; ++x) {
				int i = x + y * SEG_X;
				xoffset[i] = x * w;
				yoffset[i] = y * h;
				
				if (x < SEG_X-1)
					width[i] = w;
				else
					width[i] = canvasWidth - xoffset[i];
				
				if (y < SEG_Y-1)
					height[i] = h;
				else
					height[i] = canvasHeight - yoffset[i];
			}
		}

		for (int i = 0; i < NUM_JOBS; ++i) {
			samples[i] = new double[width[i]][height[i]][3];
		}
		
		buffer = new BufferedImage(canvasWidth, canvasHeight,
				BufferedImage.TYPE_INT_RGB);
	}
	
	public synchronized void updateCanvas(String warningText) {
		if (canvas.isDisplayable()) {
			try {
				Graphics g = canvas.getGraphics();
				g.drawImage(buffer,
						0, 0, canvas.getWidth(), canvas.getHeight(),
						0, 0, canvasWidth, canvasHeight, null);

				int width = canvas.getWidth();
				int height = canvas.getHeight();
				if (!warningText.isEmpty()) {
					g.setColor(java.awt.Color.red);
					int x0 = width/2;
					int y0 = height/2;
					g.setFont(infoFont);
					if (fontMetrics == null) {
						fontMetrics = g.getFontMetrics();
					}
					g.drawString(warningText,
							x0 - fontMetrics.stringWidth(warningText)/2, y0);
				} else {
				
					if (!bufferedScene.pathTrace()) {
						int x0 = width/2;
						int y0 = height/2;
						g.setColor(java.awt.Color.white);
						g.drawLine(x0, y0-4, x0, y0+4);
						g.drawLine(x0-4, y0, x0+4, y0);
						Ray ray = rayPool[0].get();
						bufferedScene.trace(ray);
						g.setFont(infoFont);
						if (ray.hit) {
							g.drawString(String.format("target: %.2f m", ray.distance), 5, height-18);
							g.drawString(String.format("[0x%08X] %s",
									ray.currentMaterial,
									ray.getCurrentBlock()), 5, height-5);
						}
						Vector3d pos = scene.camera().getPosition();
						g.drawString(String.format("(%.1f, %.1f, %.1f)",
								pos.x, pos.y, pos.z), 5, 11);
						rayPool[0].dispose(ray);
					}
				}
				
				/*doLensTest(g, .5, 0);
				doLensTest(g, .3, 0);
				doLensTest(g, .2, 0);
				doLensTest(g, .1, 0);
				doLensTest(g, -.1, 0);
				doLensTest(g, -.2, 0);
				doLensTest(g, -.3, 0);
				doLensTest(g, -.5, 0);*/
				
				g.dispose();
				canvas.update();
			} catch (IllegalStateException e) {
				
			}
		}
	}

	private synchronized void finalizeJob(int jobId) {
		Raster raster = buffer.getRaster();
		DataBufferInt dataBuffer = (DataBufferInt) raster.getDataBuffer();
		int[] data = dataBuffer.getData();
		double r=0, g=0, b=0;
		for (int x = 0; x < width[jobId]; ++x) {
			for (int y = 0; y < height[jobId]; ++y) {
				int xp = xoffset[jobId] + x;
				int yp = yoffset[jobId] + y;
				
				r = samples[jobId][x][y][0];
				g = samples[jobId][x][y][1];
				b = samples[jobId][x][y][2];
				
				if (bufferedScene.pathTrace()) {
					if (bufferedScene.gammaCorrection) {
						double corr = 1/bufferedScene.gamma;
						r = Math.pow(r, corr);
						g = Math.pow(g, corr);
						b = Math.pow(b, corr);
					}
				} else {
					r = Math.sqrt(r);
					g = Math.sqrt(g);
					b = Math.sqrt(b);
				}
				
				r = Math.min(1, r);
				g = Math.min(1, g);
				b = Math.min(1, b);
				
				data[xp + yp * canvasWidth] = Color.getRGB(r, g, b);
			}
		}
	}
	
	private synchronized void waitOnWorkers() throws InterruptedException {
		long updated = 0;
		while (true) {
			for (int jobId = 0; jobId < NUM_JOBS; ++jobId) {
				long ticket = 1L<<jobId;
				if ((done & ticket) != 0 && (updated & ticket) == 0) {
					updated = updated | ticket;
					finalizeJob(jobId);
				}
			}

			if (done != (1L<<NUM_JOBS) - 1)
				wait();
			else
				break;
		}
	}
	
	private synchronized void giveTickets() {
		bufferedScene.gammaCorrection = scene.gammaCorrection;
		bufferedScene.gamma = scene.gamma;
		
		tickets = (1L<<NUM_JOBS) - 1;
		done = 0;
		notifyAll();
	}

	public final synchronized int getTicket() throws InterruptedException {
		while (true) {
			if (tickets != 0) {
				for (int jobId = 0; jobId < NUM_JOBS; ++jobId) {
					long ticket = 1L<<jobId;
					if ((tickets & ticket) != 0) {
						tickets = tickets ^ ticket;
						return jobId;
					}
				}
			}
		
			wait();
		}
	}
	
	public final synchronized void jobDone(int workerId) {
		done = done | (1L<<workerId);
		notifyAll();
	}
	
	public final void work(int workerId, int jobId) {

		Vector3d d = vectorPool[workerId].get();
		Vector3d ro = vectorPool[workerId].get();
		Vector3d sample = vectorPool[workerId].get();
		double aspect = canvasWidth / (double) canvasHeight;
		
		Ray ray = rayPool[workerId].get();

		if (bufferedScene.pathTrace()) {
			
			for (int x = 0; x < width[jobId]; ++x) {
				for (int y = 0; y < height[jobId]; ++y) {
					sample.set(0, 0, 0);
					for (int i = 0; i < SPP; ++i) {
						double oy = 2 * random[workerId].nextDouble();
						double ox = 2 * random[workerId].nextDouble();
						if (oy < 1)
							oy = Math.sqrt(oy) - 1;
						else
							oy = 1 - Math.sqrt(2 - oy);
						if (ox < 1)
							ox = Math.sqrt(ox) - 1;
						else
							ox = 1 - Math.sqrt(2 - ox);
						
						bufferedScene.camera().calcViewRay(ray, d, ro,
								random[numThreads], aspect,
								(.5 - (x + xoffset[jobId] + ox) / canvasWidth),
								(-.5 + (y + yoffset[jobId] + oy) / canvasHeight));
						
						bufferedScene.pathTrace(ray,
								rayPool[workerId], vectorPool[workerId],
								random[workerId]);

						sample.x += ray.color.x;
						sample.y += ray.color.y;
						sample.z += ray.color.z;
					}
					double sinv = 1.0 / (numSamples + SPP);
					samples[jobId][x][y][0] = (samples[jobId][x][y][0] *
							numSamples + sample.x) * sinv;
					samples[jobId][x][y][1] = (samples[jobId][x][y][1] *
							numSamples + sample.y) * sinv;
					samples[jobId][x][y][2] = (samples[jobId][x][y][2] *
							numSamples + sample.z) * sinv;
				}
			}
				
		} else {
			

			boolean needSecond = bufferedScene.previewCount > 1;
			for (int x = 0; x < width[jobId]; ++x) {
				double rayx = bufferedScene.camera().fovTan * aspect *
						(.5 - (double) (x + xoffset[jobId]) / canvasWidth);
				for (int y = 0; y < height[jobId]; ++y) {
					if (!needSecond && ((x+y)%2 == 0)) {
						continue;
					} else if (needSecond) {
						if ((x+y)%2 == 1) {
							if (y%2 == 0) {
								samples[jobId][x][y][0] = samples[jobId][x-1][y][0];
								samples[jobId][x][y][1] = samples[jobId][x-1][y][1];
								samples[jobId][x][y][2] = samples[jobId][x-1][y][2];
							}
							continue;
						}
					} 
					ray.setDefault();
					ray.d.set(bufferedScene.camera().fovTan *
							(-.5 + (double) (y + yoffset[jobId]) / canvasHeight),
							-1, rayx);
					ray.d.normalize();
					bufferedScene.camera().transform(ray.d);
					
					ray.x.set(bufferedScene.camera().getPosition());
					bufferedScene.quickTrace(ray);
						
					samples[jobId][x][y][0] = ray.color.x;
					samples[jobId][x][y][1] = ray.color.y;
					samples[jobId][x][y][2] = ray.color.z;
					if (needSecond && y%2 == 1) {
						if (x > 0) {
							samples[jobId][x-1][y][0] = ray.color.x;
							samples[jobId][x-1][y][1] = ray.color.y;
							samples[jobId][x-1][y][2] = ray.color.z;
						}
						if (x+2 == width[jobId]) {
							samples[jobId][x+1][y][0] = ray.color.x;
							samples[jobId][x+1][y][1] = ray.color.y;
							samples[jobId][x+1][y][2] = ray.color.z;
						}
					}
				}
			}

		}
		
		rayPool[workerId].dispose(ray);
		vectorPool[workerId].dispose(d);
		vectorPool[workerId].dispose(ro);
		vectorPool[workerId].dispose(sample);
	}

	@SuppressWarnings("unused")
	private void doLensTest(Graphics g, double x, double z) {
		double radius = .05;
		double lensOffset = .02;
		double lensD = radius;
		double offset2 = lensOffset + lensD;
		double filmWidth = .035;
		double c1 = -lensOffset - radius;
		double c2 = -offset2 + radius;
		double n1 = 1.5;
		double aperture_w = 0.03;
		double rx;
		double rz;
		double B, C, discSqrt;
		double t0, t1, t2;
		double n1n2, radicand;
		
		int w = canvas.getWidth()-1;
		int h = canvas.getHeight()-1;
		double scale = 2;
		
		Vector3d o = vectorPool[0].get();
		Vector3d n = vectorPool[0].get();
		Vector3d i = vectorPool[0].get();
		Vector3d i0 = vectorPool[0].get();
		Vector3d d = vectorPool[0].get();
		Vector3d d2 = vectorPool[0].get();
		
		while (true) {
			rx = 2 * random[numThreads].nextDouble() - 1;
			rz = 2 * random[numThreads].nextDouble() - 1;
			double s = rx*rx + rz*rz;
			if (s > Ray.EPSILON && s <= 1) {
				s = Math.sqrt(s);
				rx *= aperture_w;
				rz *= aperture_w;
				break;
			}
		}

		i0.set(filmWidth*x, 0, filmWidth*z);
		d.set(rx - i0.x, -lensOffset, - i0.z);
		d.normalize();
		
		g.setColor(java.awt.Color.green);
		g.drawArc((int) Math.round(lensOffset * scale * w), h/2 - (int) Math.round(radius * scale * h),
				(int) Math.round(2*radius * scale * w), (int) Math.round(2*radius * scale * h), 90, 180);
		g.drawArc((int) Math.round((offset2 - 2*radius) * scale * w), h/2 - (int) Math.round(radius * scale * h),
				(int) Math.round(2*radius * scale * w), (int) Math.round(2*radius * scale * h), 270, 180);
		
		// BACK surface
		o.set(i0.x, i0.y - c1, i0.z);
		
		B = 2 * d.dot(o);
		C = o.dot(o) - radius*radius;
		radicand = B*B - 4*C;
		if (radicand < Ray.EPSILON)
			return;
		discSqrt = Math.sqrt(radicand);
		if (B < 0)
			t0 = (-B - discSqrt) / 2;
		else
			t0 = (-B + discSqrt) / 2;
		
		if (t0 <= 0 || t0 == Double.POSITIVE_INFINITY) {
			return;
		}
		
		i.x = i0.x + d.x*t0;
		i.y = i0.y + d.y*t0;
		i.z = i0.z + d.z*t0;
		
		g.setColor(java.awt.Color.red);
		g.drawLine((int) Math.round(-i0.y * scale * w), h/2 + (int) Math.round(i0.x * scale * h),
				(int) Math.round(-i.y * scale * w), h/2 + (int) Math.round(i.x * scale * h));
		
		n.set(i.x, i.y - c1, i.z);
		n.normalize();
		
		n1n2 = 1 / n1;
		t1 = -n.dot(d);
		radicand = 1 - n1n2*n1n2 * (1 - t1*t1);
		if (radicand < Ray.EPSILON) {
			return;
		}
		t2 = Math.sqrt(radicand);
		if (t1 > 0) {
			d2.x = n1n2*d.x + (n1n2*t1 - t2)*n.x;
			d2.y = n1n2*d.y + (n1n2*t1 - t2)*n.y;
			d2.z = n1n2*d.z + (n1n2*t1 - t2)*n.z;
		} else {
			d2.x = n1n2*d.x - (n1n2*t1 - t2)*n.x;
			d2.y = n1n2*d.y - (n1n2*t1 - t2)*n.y;
			d2.z = n1n2*d.z - (n1n2*t1 - t2)*n.z;
		}
		
		d2.normalize();
		d.set(d2);
		
		// FRONT surface
		i0.set(i);
		o.set(i0.x, i0.y - c2, i0.z);
		
		B = 2 * d.dot(o);
		C = o.dot(o) - radius*radius;
		radicand = B*B - 4*C;
		if (radicand < Ray.EPSILON)
			return;
		discSqrt = Math.sqrt(radicand);
		if (B < 0)
			t0 = (-B - discSqrt) / 2;
		else
			t0 = (-B + discSqrt) / 2;
		
		if (t0 <= 0 || t0 == Double.POSITIVE_INFINITY) {
			return;
		}
		
		i.x = i0.x + d.x*t0;
		i.y = i0.y + d.y*t0;
		i.z = i0.z + d.z*t0;
		
		g.setColor(java.awt.Color.red);
		g.drawLine((int) Math.round(-i0.y * scale * w), h/2 + (int) Math.round(i0.x * scale * h),
				(int) Math.round(-i.y * scale * w), h/2 + (int) Math.round(i.x * scale * h));
		
		n.set(-i.x, c2 - i.y, -i.z);
		n.normalize();
		
		n1n2 = 1.5;
		t1 = -n.dot(d);
		radicand = 1 - n1n2*n1n2 * (1 - t1*t1);
		if (radicand < Ray.EPSILON) {
			return;
		}
		t2 = Math.sqrt(radicand);
		if (t1 > 0) {
			d2.x = n1n2*d.x + (n1n2*t1 - t2)*n.x;
			d2.y = n1n2*d.y + (n1n2*t1 - t2)*n.y;
			d2.z = n1n2*d.z + (n1n2*t1 - t2)*n.z;
		} else {
			d2.x = n1n2*d.x - (n1n2*t1 - t2)*n.x;
			d2.y = n1n2*d.y - (n1n2*t1 - t2)*n.y;
			d2.z = n1n2*d.z - (n1n2*t1 - t2)*n.z;
		}
		
		d2.normalize();
		d.set(d2);
		
		t0 = 2;
		i0.set(i);
		i.x = i0.x + d.x*t0;
		i.y = i0.y + d.y*t0;
		i.z = i0.z + d.z*t0;
		
		g.setColor(java.awt.Color.red);
		g.drawLine((int) Math.round(-i0.y * scale * w), h/2 + (int) Math.round(i0.x * scale * h),
				(int) Math.round(-i.y * scale * w), h/2 + (int) Math.round(i.x * scale * h));
		
		vectorPool[0].dispose(o);
		vectorPool[0].dispose(n);
		vectorPool[0].dispose(i0);
		vectorPool[0].dispose(i);
		vectorPool[0].dispose(d);
		vectorPool[0].dispose(d2);
	}
	
	public void saveScene(File targetFile)
		throws IOException, InterruptedException {
		
		sceneLock.lock();
		synchronized (this) {
			waitOnWorkers();
			
			status.println("Saving scene " + targetFile.getName());
			status.flush();
	
			bufferedScene.saveSceneDescription(targetFile);
			
			status.println("Scene saved");
			status.flush();
		}
		sceneLock.unlock();
	}
	
	public void loadScene(RenderControls controls, File targetFile)
			throws IOException, SceneLoadingError, InterruptedException {
		
		sceneLock.lock();
		synchronized (this) {
			waitOnWorkers();
	
			statusBuffer.reset();
			updateCanvas("Loading saved dump...");
			scene.loadSceneDescription(controls, targetFile);
			bufferedScene.set(scene);
			updateCanvas();
		}
		sceneLock.unlock();
	}
	
	public synchronized void loadChunks(Collection<ChunkPosition> chunksToLoad, Scene scene) {
		canvas.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		status.println("Loading chunks...");
		status.flush();
		scene.loadChunks(chunksToLoad);
		canvas.setCursor(Cursor.getDefaultCursor());
	}

	/**
	 * Save the current frame as a PNG image.
	 */
	public synchronized void saveFrame() {
		
		CenteredFileDialog fileDialog =
				new CenteredFileDialog(null, "Save Current Frame", FileDialog.SAVE);
		fileDialog.setDirectory(System.getProperty("user.dir"));
		fileDialog.setFile(bufferedScene.getSceneName()+".png");
		fileDialog.setFilenameFilter(
				new FilenameFilter() {
					@Override
					public boolean accept(File dir, String name) {
						return name.toLowerCase().endsWith(".png");
					}
				});
		fileDialog.setVisible(true);
		File targetFile = fileDialog.getSelectedFile(".png");
		if (targetFile != null) {
			if (!targetFile.getName().endsWith(".png"))
				targetFile = new File(targetFile.getPath()+".png");
			if (targetFile.exists()) {
				Object[] options = {Messages.getString("Chunky.Cancel_lbl"), //$NON-NLS-1$
						Messages.getString("Chunky.AcceptOverwrite_lbl")}; //$NON-NLS-1$
				int n = JOptionPane.showOptionDialog(null,
						String.format(Messages.getString("Chunky.Confirm_overwrite_msg"), //$NON-NLS-1$
								targetFile.getName()),
						Messages.getString("Chunky.Confirm_overwrite_title"), //$NON-NLS-1$
						JOptionPane.YES_NO_OPTION,
						JOptionPane.WARNING_MESSAGE,
						null,
						options,
						options[0]);
				if (n != 1)
					return;
			}
			try {
				addWatermark();
				ImageIO.write(buffer, "png", targetFile);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * Wait until the render controller is ready to start rendering.
	 * @throws InterruptedException 
	 */
	public synchronized void waitReady() throws InterruptedException {
		while (!isReady)
			wait();
	}

	public synchronized void setCanvasSize(int width, int height)
		throws InterruptedException {

		waitOnWorkers();
		canvasWidth = Math.max(MIN_CANVAS_WIDTH, width);
		canvasHeight = Math.max(MIN_CANVAS_HEIGHT, height);
		initBuffers();
		canvas.setPreferredSize(new Dimension(width, height));
		view.pack();
		scene.refresh();
		
		ProgramProperties.setProperty("3dcanvas.width", "" + canvasWidth);
		ProgramProperties.setProperty("3dcanvas.height", "" + canvasHeight);
	}

	public synchronized String getStatusText() throws InterruptedException {
		while (!statusTextUpdated)
			wait();
		statusTextUpdated = false;
		return statusText;
	}

	public int getCanvasWidth() {
		return canvasWidth;
	}

	public int getCanvasHeight() {
		return canvasHeight;
	}
	
	public Scene scene() {
		return scene;
	}

	public synchronized void halveCanvasSize() throws InterruptedException {
		setCanvasSize(canvasWidth/2, canvasHeight/2);
	}
	
	public void doubleCanvasSize() throws InterruptedException {
		setCanvasSize(canvasWidth*2, canvasHeight*2);
	}
	
	public void addRenderStateListener(RenderStateListener listener) {
		listeners.add(listener);
	}
	
	public void removeRenderStateListener(RenderStateListener listener) {
		listeners.remove(listener);
	}
	
	public void fireRenderStateEvent(boolean pathTrace, boolean paused) {
		for (RenderStateListener listener : listeners) {
			listener.renderStateChanged(pathTrace, paused);
		}
	}

	public void updateCanvas() {
		if (sceneLock.tryLock()) {
			synchronized (this) {
				updateCanvas(scene.loadedChunks().isEmpty() ? "No chunks loaded!" : "");
			}
			sceneLock.unlock();
		}
	}

	public void appendStatusText(String text) {
		status.append(text);
		status.flush();
	}
}
