/* Copyright (c) 2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.renderer;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import javax.swing.JOptionPane;

import se.llbit.chunky.main.SceneLoadingError;
import se.llbit.chunky.model.WaterModel;
import se.llbit.chunky.renderer.ui.RenderControls;
import se.llbit.chunky.world.Biomes;
import se.llbit.chunky.world.Block;
import se.llbit.chunky.world.BlockType;
import se.llbit.chunky.world.Chunk;
import se.llbit.chunky.world.ChunkPosition;
import se.llbit.chunky.world.Texture;
import se.llbit.chunky.world.World;
import se.llbit.math.Octree;
import se.llbit.math.QuickMath;
import se.llbit.math.Ray;
import se.llbit.math.Ray.RayPool;
import se.llbit.math.Vector3d;
import se.llbit.math.Vector3i;
import se.llbit.nbt.AnyTag;
import se.llbit.nbt.ByteTag;
import se.llbit.nbt.CompoundTag;
import se.llbit.nbt.DoubleTag;
import se.llbit.nbt.IntTag;
import se.llbit.nbt.ListTag;
import se.llbit.nbt.LongTag;
import se.llbit.nbt.NamedTag;
import se.llbit.nbt.StringTag;
import se.llbit.nbt.Tag;
import se.llbit.util.ProgramProperties;
import se.llbit.util.VectorPool;

/**
 * Scene description.
 */
public class Scene {
	private static final float SPECULAR_COEFF = 0.31f;
	public static final float WATER_SPECULAR = 0.46f;
	public static final double MIN_GAMMA = 1;
	public static final double MAX_GAMMA = 4;
	public static final double DEFAULT_GAMMA = 2.2;
	public static final double DEFAULT_EMITTER_INTENSITY = 13;
	
	public static final int CVF_VERSION = 1;
	
	private static final double WATER_FOG_DEPTH = 8.5;

	public static int rayDepth = 5;
	
	private String name = "default";
	private String skymapFileName;

	private final Sky sky = new Sky(this);
	private final Camera camera = new Camera(this);
	private final Sun sun = new Sun(this);
	
	/**
 	 * World
 	 */
	private World world;

	/**
 	 * Octree origin
 	 */
	private Vector3i origin = new Vector3i();
	
	/**
 	 * Octree
 	 */
	private Octree octree;

	double gamma = DEFAULT_GAMMA;
	private boolean stillWater = false;
	private boolean biomeColors = true;
	private boolean sunEnabled = true;
	private boolean emittersEnabled = true;

	private Set<ChunkPosition> loadedChunks = new HashSet<ChunkPosition>();

	// chunk loading buffers
	private byte[] blocks = new byte[Chunk.X_MAX * Chunk.Y_MAX * Chunk.Z_MAX];
	private byte[] biomes = new byte[Chunk.X_MAX * Chunk.Z_MAX];
	private byte[] data = new byte[(Chunk.X_MAX * Chunk.Y_MAX * Chunk.Z_MAX) / 2];

	private boolean refresh = false;
	private boolean pathTrace = false;
	boolean pauseRender = true;
	
	/**
 	 * Preview frame interlacing counter.
 	 */
	int previewCount = 2;
	
	private RenderManager controller;
	private boolean clearWater = false;
	public boolean gammaCorrection = true;
	private double emitterIntensity = DEFAULT_EMITTER_INTENSITY;
	private boolean atmosphereEnabled = false;
	private boolean volumetricFogEnabled = false;
	
	public Scene(RenderManager controller, World world) {
		this.controller = controller;
		this.world = world;
		octree = new Octree(1);// empty octree
		name = world.levelName();
		
		if (ProgramProperties.getProperty("skymap") != null)
			sky.loadSkyMap(ProgramProperties.getProperty("skymap"));
	}

	public Scene(Scene other) {
		this.controller = other.controller;
		set(other);
	}

	public void set(Scene other) {
		world = other.world;

		// the octree reference is overwritten to save time
		// when the other scene is changed it must create a new octree
		// TODO: appendable octrees?
		octree = other.octree;
		origin.set(other.origin);
		
		loadedChunks = other.loadedChunks;

		gamma = other.gamma;
		name = other.name;
		skymapFileName = other.skymapFileName;
		
		stillWater = other.stillWater;
		clearWater = other.clearWater;
		biomeColors = other.biomeColors;
		sunEnabled = other.sunEnabled;
		emittersEnabled = other.emittersEnabled;
		emitterIntensity = other.emitterIntensity;
		atmosphereEnabled = other.atmosphereEnabled;
		volumetricFogEnabled = other.volumetricFogEnabled;
		
		previewCount = 2;
		pathTrace = other.pathTrace;
		pauseRender = other.pauseRender;
		
		camera.set(other.camera);
		sky.set(other.sky);
		sun.set(other.sun);
		
		WaterModel.stillWater = stillWater;
		Biomes.biomeColorsEnabled = biomeColors;
	}
	
	private String getFrameName(String fileName) {
		return fileName.split("\\.")[0];
	}

	public synchronized void saveSceneDescription(File targetFile)
		throws IOException, InterruptedException {
		
		name = getFrameName(targetFile.getName());
		DataOutputStream out = new DataOutputStream(
				new FileOutputStream(targetFile));
		
		CompoundTag chunkView = new CompoundTag();
		chunkView.addItem(new NamedTag("version", new IntTag(CVF_VERSION)));
		
		CompoundTag canvas = new CompoundTag();
		CompoundTag worldTag = new CompoundTag();
		
		canvas.addItem("width", new IntTag(controller.canvasWidth));
		canvas.addItem("height", new IntTag(controller.canvasHeight));
		canvas.addItem("gamma", new DoubleTag(gamma));
		
		worldTag.addItem("worldDirectoryPath",
				new StringTag(world.getWorldDirectory().getAbsolutePath()));
		worldTag.addItem("dimension",
				new IntTag(world.currentDimension()));
		worldTag.addItem("emittersEnabled",
				new ByteTag(emittersEnabled ? 1 : 0));
		worldTag.addItem("sunEnabled",
				new ByteTag(sunEnabled ? 1 : 0));
		worldTag.addItem("stillWater",
				new ByteTag(stillWater ? 1 : 0));
		worldTag.addItem("clearWater",
				new ByteTag(clearWater ? 1 : 0));
		worldTag.addItem("biomeColorsEnabled",
				new ByteTag(biomeColors ? 1 : 0));
		worldTag.addItem("atmosphereEnabled",
				new ByteTag(atmosphereEnabled ? 1 : 0));
		worldTag.addItem("volumetricFogEnabled",
				new ByteTag(volumetricFogEnabled ? 1 : 0));
		worldTag.addItem("emitterIntensity", new DoubleTag(emitterIntensity));
		sky.save(worldTag);

		ListTag chunkList = new ListTag();
		chunkList.setType(Tag.TAG_LONG);
		for (ChunkPosition cp : loadedChunks) {
			chunkList.addItem(new LongTag(cp.getLong()));
		}
		
		chunkView.addItem("canvas", canvas);
		chunkView.addItem("camera", camera.store());
		chunkView.addItem("sun", sun.store());
		chunkView.addItem("world", worldTag);
		chunkView.addItem("chunkList", chunkList);
		
		NamedTag rootTag = new NamedTag("chunkView", chunkView);
		rootTag.write(out);
		out.close();
		
		controller.saveDump(name + ".dump");
	}
	
	public synchronized void loadSceneDescription(RenderControls controls, File targetFile)
			throws IOException, SceneLoadingError, InterruptedException {
		
		DataInputStream in = new DataInputStream(
				new FileInputStream(targetFile));
		
		String cvf_ver = "chunkView.version";
		String cvf_canvas = "chunkView.canvas";
		String cvf_camera = "chunkView.camera";
		String cvf_world = "chunkView.world";
		String cvf_sun = "chunkView.sun";
		String cvf_chunkList = "chunkView.chunkList";
		
		Set<String> request = new HashSet<String>();
		request.add(cvf_ver);
		request.add(cvf_canvas);
		request.add(cvf_camera);
		request.add(cvf_world);
		request.add(cvf_sun);
		request.add(cvf_chunkList);
		Map<String, AnyTag> result = NamedTag.quickParse(in, request);
		in.close();
		
		if (!result.containsKey(cvf_ver) ||
				result.get(cvf_ver).isError() ||
				result.get(cvf_ver).intValue() != CVF_VERSION) {
			
			if (result.containsKey(cvf_ver) &&
					!result.get(cvf_ver).isError())
				throw new SceneLoadingError("Incorrect CVF version! " +
						"Expected " + CVF_VERSION + " but fund " +
						result.get(cvf_ver).intValue() + ".");
			else
				throw new SceneLoadingError("Could not read CVF version!");
		}
		
		name = getFrameName(targetFile.getName());
		
		if (result.containsKey(cvf_canvas)) {
			CompoundTag tag = (CompoundTag) result.get(cvf_canvas);
			
			int width = tag.get("width").intValue();
			int height = tag.get("height").intValue();
			controller.setCanvasSize(width, height);
			gamma = tag.get("gamma").doubleValue(DEFAULT_GAMMA);
		}
		
		if (result.containsKey(cvf_camera))  {
			
			camera.load((CompoundTag) result.get(cvf_camera));
		}
		
		if (result.containsKey(cvf_world))  {
			CompoundTag worldTag = (CompoundTag) result.get(cvf_world);
			
			String worldDirectoryPath = worldTag.get("worldDirectoryPath").stringValue();
			int dimension = worldTag.get("dimension").intValue();
			if (!worldDirectoryPath.isEmpty()) {
				File worldDirectory = new File(worldDirectoryPath);
				if (worldDirectory.exists() && worldDirectory.isDirectory()) {
					File levelDat = new File(worldDirectory, "level.dat");
					if (!levelDat.exists()) {
						JOptionPane.showMessageDialog(null,
								"Could not load world: " + worldDirectoryPath);
					} else {
						if (world.getWorldDirectory() == null ||
								!world.getWorldDirectory().getAbsolutePath().equals(worldDirectoryPath)) {
							world = new World(worldDirectory);
							world.setDimension(dimension);
						} else if (world.currentDimension() != dimension) {
							world.setDimension(dimension);
						}
					}
				}
			}
			emittersEnabled =
				worldTag.get("emittersEnabled").byteValue() != 0;
			sunEnabled = worldTag.get("sunEnabled").byteValue() != 0;
			stillWater = worldTag.get("stillWater").byteValue() != 0;
			clearWater = worldTag.get("clearWater").byteValue() != 0;
			biomeColors =
				worldTag.get("biomeColorsEnabled").byteValue() != 0;
			atmosphereEnabled = worldTag.get("atmosphereEnabled").byteValue() != 0;
			volumetricFogEnabled = worldTag.get("volumetricFogEnabled").byteValue() != 0;
			emitterIntensity = worldTag.get("emitterIntensity").doubleValue(DEFAULT_EMITTER_INTENSITY);
			sky.load(worldTag);
		}
		
		if (result.containsKey(cvf_sun))  {
			
			sun.load((CompoundTag) result.get(cvf_sun));
		}

		pathTrace = true;
		pauseRender = true;
		refresh = false;
		
		controller.loadDump(name + ".dump");
		
		if (controls != null)
			controls.sceneDescriptionLoaded();
		
		if (result.containsKey(cvf_chunkList)) {
			ListTag chunkList = (ListTag) result.get(cvf_chunkList);
			
			Collection<ChunkPosition> chunksToLoad =
				new LinkedList<ChunkPosition>();
			for (int i = 0; i < chunkList.getNumItem(); ++i) {
				chunksToLoad.add(ChunkPosition.get(
							chunkList.getItem(i).longValue()));
			}
			
			controller.loadChunks(chunksToLoad, this);
		}
	}

	public synchronized void setGamma(double value) {
		gamma = value;
		if (!pathTrace)
			refresh();
	}

	public double getGamma() {
		return gamma;
	}

	public void setStillWater(boolean value) {
		if (value != stillWater) {
			stillWater = value;
			refresh();
		}
	}

	public String getSceneName() {
		return name;
	}

	public boolean getDirectLight() {
		return sunEnabled;
	}

	public synchronized void setEmittersEnabled(boolean value) {
		if (value != emittersEnabled) {
			emittersEnabled = value;
			refresh();
		}
	}

	public synchronized void setDirectLight(boolean value) {
		if (value != sunEnabled) {
			sunEnabled = value;
			refresh();
		}
	}
	
	public boolean getEmittersEnabled() {
		return emittersEnabled;
	}

	/**
	 * Test whether the ray intersects any voxel before exiting the Octree.
	 * @param ray the ray
	 * @return <code>true</code> if the ray intersects a voxel
	 */
	public boolean intersect(Ray ray) {

		int level;
		Octree.Node node;
		boolean first = true;

		int lx, ly, lz;
		int x, y, z;
		int nx = 0, ny = 0, nz = 0;
		double tNear = Double.POSITIVE_INFINITY;
		double t;
		
		while (true) {
			Vector3d d = ray.d;
			
			// add small offset past the intersection to avoid
			// recursion to the same octree node!
			x = QuickMath.floor(ray.x.x + ray.d.x * Ray.OFFSET);
			y = QuickMath.floor(ray.x.y + ray.d.y * Ray.OFFSET);
			z = QuickMath.floor(ray.x.z + ray.d.z * Ray.OFFSET);
			
			node = octree.root;
			level = octree.depth;
			lx = x >>> level;
			ly = y >>> level;
			lz = z >>> level;
			
			if (lx != 0 || ly != 0 || lz != 0) {
				
				// ray origin is outside octree!
				ray.currentMaterial = Block.AIR.id;
				
				// only check octree intersection if this is the first iteration
				if (first) {
					// test if it is entering the octree
					t = -ray.x.x / d.x;
					if (t > Ray.EPSILON) {
						tNear = t;
						nx = 1;
						ny = nz = 0;
					}
					t = ((1<<level) - ray.x.x) / d.x;
					if (t < tNear && t > Ray.EPSILON) {
						tNear = t;
						nx = -1;
						ny = nz = 0;
					}
					t = -ray.x.y / d.y;
					if (t < tNear && t > Ray.EPSILON) {
						tNear = t;
						ny = 1;
						nx = nz = 0;
					}
					t = ((1<<level) - ray.x.y) / d.y;
					if (t < tNear && t > Ray.EPSILON) {
						tNear = t;
						ny = -1;
						nx = nz = 0;
					}
					t = -ray.x.z / d.z;
					if (t < tNear && t > Ray.EPSILON) {
						tNear = t;
						nz = 1;
						nx = ny = 0;
					}
					t = ((1<<level) - ray.x.z) / d.z;
					if (t < tNear && t > Ray.EPSILON) {
						tNear = t;
						nz = -1;
						nx = ny = 0;
					}
					
					if (tNear < Double.MAX_VALUE) {
						ray.x.scaleAdd(tNear, d, ray.x);
						ray.n.set(nx, ny, nz);
						ray.distance += tNear;
						tNear = Double.POSITIVE_INFINITY;
						continue;
					} else {
						return false;// outside of octree!
					}
				} else {
					return false;// outside of octree!
				}
			}
			
			first = false;
			
			while (node.type == -1) {
				level -= 1;
				lx = x >>> level;
				ly = y >>> level;
				lz = z >>> level;
				node = node.children[((lx&1)<<2) | ((ly&1)<<1) | (lz&1)];
			}
			
			/*double w = .1 * (1 + level);
			w*=w;
			if (ray.x.x < (lx<<level) + w && (ray.x.y < (ly<<level) + w || ray.x.y > ((ly+1)<<level) - w) ||
					ray.x.x < (lx<<level) + w && (ray.x.z < (lz<<level) + w || ray.x.z > ((lz+1)<<level) - w) ||
					ray.x.y < (ly<<level) + w && (ray.x.z < (lz<<level) + w || ray.x.z > ((lz+1)<<level) - w) ||
					ray.x.x > ((lx+1)<<level) - w && (ray.x.y < (ly<<level) + w || ray.x.y > ((ly+1)<<level) - w) ||
					ray.x.x > ((lx+1)<<level) - w && (ray.x.z < (lz<<level) + w || ray.x.z > ((lz+1)<<level) - w) ||
					ray.x.y > ((ly+1)<<level) - w && (ray.x.z < (lz<<level) + w || ray.x.z > ((lz+1)<<level) - w)) {
				ray.color.x = .5;
				ray.color.y = .5;
				ray.color.z = .5;
				ray.color.w = 1;
				ray.prevMaterial = Block.AIR.id;
				ray.currentMaterial = 0xFF;
				return true;
			}*/

			if (ray.currentMaterial == -1) {
				ray.prevMaterial = 0;
				ray.currentMaterial = node.type;
			}

			Block currentBlock = Block.values[node.type & 0xFF];
			Block prevBlock = Block.values[ray.currentMaterial & 0xFF];

			if ((currentBlock != Block.WATER || prevBlock != Block.WATER) &&
					(currentBlock != Block.LAVA || prevBlock != Block.LAVA) &&
					currentBlock.localIntersect) {

				ray.prevMaterial = ray.currentMaterial;
				ray.currentMaterial = node.type;
				
				if (currentBlock.intersect(ray)) {

					return true;

				} else {
					// exit ray from this local block
					ray.currentMaterial = 0;// current material is air
					
					ray.exitBlock(x, y, z);
					continue;
				}
			}

			if (currentBlock != prevBlock) {
				ray.prevMaterial = ray.currentMaterial;
				ray.currentMaterial = node.type;
				getIntersectionColor(ray);
				return true;
			}

			t = ((lx<<level) - ray.x.x) / d.x;
			if (t > Ray.EPSILON) {
				tNear = t;
				nx = 1;
				ny = nz = 0;
			}
			t = (((lx+1)<<level) - ray.x.x) / d.x;
			if (t < tNear && t > Ray.EPSILON) {
				tNear = t;
				nx = -1;
				ny = nz = 0;
			}
			t = ((ly<<level) - ray.x.y) / d.y;
			if (t < tNear && t > Ray.EPSILON) {
				tNear = t;
				ny = 1;
				nx = nz = 0;
			}
			t = (((ly+1)<<level) - ray.x.y) / d.y;
			if (t < tNear && t > Ray.EPSILON) {
				tNear = t;
				ny = -1;
				nx = nz = 0;
			}
			t = ((lz<<level) - ray.x.z) / d.z;
			if (t < tNear && t > Ray.EPSILON) {
				tNear = t;
				nz = 1;
				nx = ny = 0;
			}
			t = (((lz+1)<<level) - ray.x.z) / d.z;
			if (t < tNear && t > Ray.EPSILON) {
				tNear = t;
				nz = -1;
				nx = ny = 0;
			}

			ray.x.scaleAdd(tNear, d, ray.x);
			ray.n.set(nx, ny, nz);
			ray.distance += tNear;
			tNear = Double.POSITIVE_INFINITY;
		}
	}

	/**
	 * @param ray
	 */
	public void quickTrace(Ray ray) {
		
		ray.x.x -= origin.x;
		ray.x.y -= origin.y;
		ray.x.z -= origin.z;

		while (true) {
			if (!intersect(ray)) {
				if (controller.waterHeight > 0 &&
						ray.d.y < 0 && ray.x.y > controller.waterHeight-.125) {
					
					ray.t = (controller.waterHeight-.125-ray.x.y) / ray.d.y;
					ray.distance += ray.t;
					ray.x.scaleAdd(ray.t, ray.d, ray.x);
					ray.currentMaterial = Block.WATER.id;
					ray.prevMaterial = 0;
					WaterModel.intersect(ray);
				}
				break;
			} else if (ray.currentMaterial != 0 && ray.color.w > 0) {
				break;
			} else {
				ray.t += Ray.OFFSET;
				ray.x.scaleAdd(Ray.OFFSET, ray.d, ray.x);
			}
		}
		
		if (ray.currentMaterial == 0) {
			sky.getSkyColor4(ray, false);
		} else {
			sun.flatShading(ray);
		}
	}

	public void pathTrace(Ray ray, RayPool pool,
			VectorPool vectorPool, Random random) {

		ray.x.x -= origin.x;
		ray.x.y -= origin.y;
		ray.x.z -= origin.z;
		
		pathTrace(ray, pool, vectorPool, random, 1, true);
	}

	private void getIntersectionColor(Ray ray) {

		if (ray.currentMaterial == 0) {
			ray.color.x = 1;
			ray.color.y = 1;
			ray.color.z = 1;
			ray.color.w = 0;
			return;
		}

		int x = QuickMath.floor(ray.x.x);
		int y = QuickMath.floor(ray.x.y);
		int z = QuickMath.floor(ray.x.z);
		ray.calcUVCoords(x, y, z);

		Block block = ray.getCurrentBlock();
		switch (block) {
		case WOODENPLANKS:
			Texture.woodPlank[ray.getBlockMetadata() % 4].
				getColor(ray.u, ray.v, ray.color);
			break;
		case WOOL:
			Texture.wool[ray.getBlockMetadata()].
				getColor(ray.u, ray.v, ray.color);
			break;
		default:
			block.getTexture().getColor(ray.u, ray.v, ray.color);
		}
	}
	
	public void pathTrace(Ray ray, RayPool rayPool,
			VectorPool vectorPool, Random random, int addEmitted,
			boolean first) {
		
		Ray reflected = rayPool.get();
		Ray transmitted = rayPool.get();
		Ray refracted = rayPool.get();
		Vector3d ox = vectorPool.get(ray.x);
		Vector3d od = vectorPool.get(ray.d);
		double s = 0;
		
		while (true) {

			if (!intersect(ray)) {
				if (controller.waterHeight > 0 &&
						ray.d.y < 0 && ray.x.y > controller.waterHeight-.125) {
					
					ray.t = (controller.waterHeight-.125-ray.x.y) / ray.d.y;
					ray.distance += ray.t;
					ray.x.scaleAdd(ray.t, ray.d, ray.x);
					ray.currentMaterial = Block.WATER.id;
					ray.prevMaterial = 0;
					WaterModel.intersect(ray);
					
					if (first) {
						s = ray.distance;
						first = false;
					}
				} else {
					if (ray.depth == 0) {
						// direct sky hit
						sky.getSkyPanoramaColorInterpolated(ray, controller.waterHeight > 0);
						
					} else if (ray.specular) {
						// sky color
						sky.getSkyColor4(ray, controller.waterHeight > 0);
					} else {
						sky.getSkyPanoramaColor(ray, controller.waterHeight > 0);
					}
					break;
				}
			}

			double pSpecular = 0;

			Block currentBlock = ray.getCurrentBlock();
			if (BlockType.shiny.contains(currentBlock)) {
				if (currentBlock == Block.WATER)
					pSpecular = WATER_SPECULAR;
				else
					pSpecular = SPECULAR_COEFF;
			}

			double pDiffuse = ray.color.w;
			
			float n1 = ray.getPrevBlock().refractionIndex;
			float n2 = ray.getCurrentBlock().refractionIndex;
			
			if (pDiffuse + pSpecular < Ray.EPSILON && n1 == n2)
				continue;
			
			if (first) {
				s = ray.distance;
				first = false;
			}
			
			if (BlockType.shiny.contains(currentBlock) &&
					random.nextDouble() < pSpecular) {

				getSpecularReflectedRay(ray, reflected);

				if (!kill(reflected, random)) {
					pathTrace(reflected, rayPool, vectorPool, random, 1, false);
					if (reflected.hit) {
						ray.color.x *= reflected.color.x;
						ray.color.y *= reflected.color.y;
						ray.color.z *= reflected.color.z;
						ray.hit = true;
					}
				}

			} else {

				if (random.nextDouble() < pDiffuse) {

					reflected.set(ray);
					if (!kill(reflected, random)) {

						double emittance = 0;

						if (emittersEnabled && BlockType.emitter.contains(currentBlock)) {

							emittance = addEmitted;
							ray.emittance.x = ray.color.x * ray.color.x * emitterIntensity;
							ray.emittance.y = ray.color.y * ray.color.y * emitterIntensity;
							ray.emittance.z = ray.color.z * ray.color.z * emitterIntensity;
							ray.hit = true;
						}

						if (sunEnabled) {
							sun.getRandomSunDirection(reflected, random, vectorPool);

							double directLight = 0;
							
							if (reflected.d.dot(ray.n) > 0) {
							
								reflected.currentMaterial = ray.prevMaterial;
		
								double attenuation = getDirectLightAttenuation(reflected);
								
								if (attenuation > 0) {
									directLight = attenuation * reflected.d.dot(ray.n);
									ray.hit = true;
								}
							}
	
							getDiffuseReflectedRay(ray, reflected, random);
							
							pathTrace(reflected, rayPool, vectorPool, random, 0, false);
							ray.hit = ray.hit || reflected.hit;
							if (ray.hit) {
								ray.color.x = ray.color.x
									* (emittance + directLight * sun.emittance.x
										+ (reflected.color.x + reflected.emittance.x));
								ray.color.y = ray.color.y
									* (emittance + directLight * sun.emittance.y
										+ (reflected.color.y + reflected.emittance.y));
								ray.color.z = ray.color.z
									* (emittance + directLight * sun.emittance.z
										+ (reflected.color.z + reflected.emittance.z));
							}
						} else {
							getDiffuseReflectedRay(ray, reflected, random);
							
							pathTrace(reflected, rayPool, vectorPool, random, 0, false);
							ray.hit = ray.hit || reflected.hit;
							if (ray.hit) {
								ray.color.x = ray.color.x
									* (emittance + (reflected.color.x + reflected.emittance.x));
								ray.color.y = ray.color.y
									* (emittance + (reflected.color.y + reflected.emittance.y));
								ray.color.z = ray.color.z
									* (emittance + (reflected.color.z + reflected.emittance.z));
							}
						}
					}
				} else if (n1 != n2) {
					
					boolean doRefraction = ray.getCurrentBlock() == Block.WATER
							|| ray.getPrevBlock() == Block.WATER
							|| ray.getCurrentBlock() == Block.ICE
							|| ray.getPrevBlock() == Block.ICE;
					
					// refraction
					float n1n2 = n1 / n2;
					double cosTheta = - ray.n.dot(ray.d);
					double radicand = 1 - n1n2*n1n2 * (1 - cosTheta*cosTheta);
					if (doRefraction && radicand < Ray.EPSILON) {
						// total internal reflection
						getSpecularReflectedRay(ray, reflected);
						if (!kill(reflected, random)) {
							pathTrace(reflected, rayPool, vectorPool, random, 1, false);
							if (reflected.hit) {
								
								ray.color.x = reflected.color.x;
								ray.color.y = reflected.color.y;
								ray.color.z = reflected.color.z;
								ray.hit = true;
							}
						}
					} else {
						refracted.set(ray);
						if (!kill(refracted, random)) {
							
							// Calculate angle-dependent reflectance using
							// Fresnel equation approximation
							// R(theta) = R0 + (1 - R0) * (1 - cos(theta))^5
							float a = (n1n2 - 1);
							float b = (n1n2 + 1);
							double R0 = a*a/(b*b);
							double c = 1 - cosTheta;
							double Rtheta = R0 + (1-R0) * c*c*c*c*c;
							
							if (random.nextDouble() < Rtheta) {
								getSpecularReflectedRay(ray, reflected);
								pathTrace(reflected, rayPool, vectorPool, random, 1, false);
								if (reflected.hit) {
									ray.color.x = reflected.color.x;
									ray.color.y = reflected.color.y;
									ray.color.z = reflected.color.z;
									ray.hit = true;
								}
							} else {
								if (doRefraction) {
									
									double t2 = Math.sqrt(radicand);
									if (cosTheta > 0) {
										refracted.d.x = n1n2*ray.d.x + (n1n2*cosTheta - t2)*ray.n.x;
										refracted.d.y = n1n2*ray.d.y + (n1n2*cosTheta - t2)*ray.n.y;
										refracted.d.z = n1n2*ray.d.z + (n1n2*cosTheta - t2)*ray.n.z;
									} else {
										refracted.d.x = n1n2*ray.d.x - (n1n2*cosTheta - t2)*ray.n.x;
										refracted.d.y = n1n2*ray.d.y - (n1n2*cosTheta - t2)*ray.n.y;
										refracted.d.z = n1n2*ray.d.z - (n1n2*cosTheta - t2)*ray.n.z;
									}
									
									refracted.d.normalize();
									
									refracted.x.scaleAdd(Ray.OFFSET,
											refracted.d, refracted.x);
								}
								
								pathTrace(refracted, rayPool, vectorPool, random, 1, false);
								if (refracted.hit) {
									ray.color.x = ray.color.x * pDiffuse + (1-pDiffuse);
									ray.color.y = ray.color.y * pDiffuse + (1-pDiffuse);
									ray.color.z = ray.color.z * pDiffuse + (1-pDiffuse);
									ray.color.x *= refracted.color.x;
									ray.color.y *= refracted.color.y;
									ray.color.z *= refracted.color.z;
									ray.hit = true;
								}
							}
						}
					}

				} else {

					transmitted.set(ray);
					transmitted.x.scaleAdd(Ray.OFFSET, transmitted.d,
							transmitted.x);

					pathTrace(transmitted, rayPool, vectorPool, random, 1, false);
					if (transmitted.hit) {
						ray.color.x = ray.color.x * pDiffuse + (1-pDiffuse);
						ray.color.y = ray.color.y * pDiffuse + (1-pDiffuse);
						ray.color.z = ray.color.z * pDiffuse + (1-pDiffuse);
						ray.color.x *= transmitted.color.x;
						ray.color.y *= transmitted.color.y;
						ray.color.z *= transmitted.color.z;
						ray.hit = true;
					}
				}
			}
			
			// do water fog
			if (!clearWater && ray.hit && ray.getPrevBlock() == Block.WATER) {
				double a = ray.distance / WATER_FOG_DEPTH;
				double attenuation = 1 - Math.min(1, a*a);
				ray.color.x *= attenuation;
				ray.color.y *= attenuation;
				ray.color.z *= attenuation;
				ray.color.w = attenuation;
			}
			
			break;
		}
		if (!ray.hit) {
			ray.color.set(0, 0, 0, 1);
			if (first)
				s = ray.distance;
		}
		
		if (s > 0) {
			
			if (atmosphereEnabled) {
				double Fex = sun.extinction(s);
				ray.color.x *= Fex;
				ray.color.y *= Fex;
				ray.color.z *= Fex;
				
				if (!volumetricFogEnabled) {
					double Fin = sun.inscatter(Fex, sun.theta(ray.d));
					
					ray.color.x += Fin * sun.emittance.x * sun.getIntensity();
					ray.color.y += Fin * sun.emittance.y * sun.getIntensity();
					ray.color.z += Fin * sun.emittance.z * sun.getIntensity();
				}
			}
			
			if (volumetricFogEnabled) {
				s = (s - Ray.OFFSET) * random.nextDouble();
				
				reflected.x.scaleAdd(s, od, ox);
				sun.getRandomSunDirection(reflected, random, vectorPool);
				reflected.currentMaterial = 0;
				
				double attenuation = getDirectLightAttenuation(reflected);
				
				double Fex = sun.extinction(s);
				double Fin = sun.inscatter(Fex, sun.theta(ray.d));
				
				ray.color.x += 50 * attenuation * Fin * sun.emittance.x * sun.getIntensity();
				ray.color.y += 50 * attenuation * Fin * sun.emittance.y * sun.getIntensity();
				ray.color.z += 50 * attenuation * Fin * sun.emittance.z * sun.getIntensity();
			}
		}
		
		rayPool.dispose(reflected);
		rayPool.dispose(transmitted);
		rayPool.dispose(refracted);
		vectorPool.dispose(ox);
		vectorPool.dispose(od);
	}

	private double getDirectLightAttenuation(Ray ray) {
		double attenuation = 1;
		while (attenuation > 0) {
			ray.x.scaleAdd(Ray.OFFSET,
					ray.d, ray.x);
			if (!intersect(ray))
				break;
			if (!clearWater && ray.getPrevBlock() == Block.WATER) {
				double a = ray.distance / WATER_FOG_DEPTH;
				attenuation *= 1 - Math.min(1, a*a);
			} else if (ray.currentMaterial != 0)
				attenuation *= 1 - ray.color.w;
		}
		return attenuation;
	}

	public static final void getDiffuseReflectedRay(Ray ray, Ray reflected, Random random) {
		reflected.set(ray);

		Vector3d d = reflected.d;
		Vector3d n = reflected.n;
		
		// get random point on unit disk
		double x1 = random.nextDouble();
		double x2 = random.nextDouble();
		double r = Math.sqrt(x1);
		double theta = 2 * Math.PI * x2;
		
		// project to point on hemisphere in tangent space
		double tx = r * Math.cos(theta);
		double ty = r * Math.sin(theta);
		double tz = Math.sqrt(1 - x1);
		
		// transform from tangent space to world space
		double xx, xy, xz;
		double ux, uy, uz;
		double vx, vy, vz;
		
		if (Math.abs(n.x) > .1) {
			xx = 0;
			xy = 1;
			xz = 0;
		} else {
			xx = 1;
			xy = 0;
			xz = 0;
		}
		
		ux = xy * n.z - xz * n.y;
		uy = xz * n.x - xx * n.z;
		uz = xx * n.y - xy * n.x;
		
		r = 1/Math.sqrt(ux*ux + uy*uy + uz*uz);
		
		ux *= r;
		uy *= r;
		uz *= r;
		
		vx = uy * n.z - uz * n.y;
		vy = uz * n.x - ux * n.z;
		vz = ux * n.y - uy * n.x;
		
		d.x = ux * tx + vx * ty + n.x * tz;
		d.y = uy * tx + vy * ty + n.y * tz;
		d.z = uz * tx + vz * ty + n.z * tz;
		d.normalize();
		
		reflected.x.scaleAdd(Ray.OFFSET, reflected.d, reflected.x);
		reflected.currentMaterial = reflected.prevMaterial;
		reflected.specular = false;
	}

	public static final void getSpecularReflectedRay(Ray ray, Ray reflected) {
		reflected.set(ray);
		reflected.d.scaleAdd(
				- 2 * ray.d.dot(ray.n),
				ray.n, ray.d);
		reflected.x.scaleAdd(Ray.OFFSET, reflected.d, reflected.x);
		reflected.currentMaterial = reflected.prevMaterial;
	}
	
	private final boolean kill(Ray ray, Random random) {
		return ray.depth >= rayDepth && random.nextDouble() < .5f;
	}

	public synchronized void loadChunks(Collection<ChunkPosition> chunksToLoad) {
		int emitters = 0;
		int fence = 0;
		int chunks = 0;
		
		if (chunksToLoad.isEmpty()) {
			return;
		}
		
		loadedChunks = new HashSet<ChunkPosition>();
		
		int xmin = Integer.MAX_VALUE;
		int xmax = Integer.MIN_VALUE;
		int zmin = Integer.MAX_VALUE;
		int zmax = Integer.MIN_VALUE;
		for (ChunkPosition cp: chunksToLoad) {
			if (cp.x < xmin)
				xmin = cp.x;
			if (cp.x > xmax)
				xmax = cp.x;
			if (cp.z < zmin)
				zmin = cp.z;
			if (cp.z > zmax)
				zmax = cp.z;
		}
		
		xmax += 1;
		zmax += 1;
		xmin *= 16;
		xmax *= 16;
		zmin *= 16;
		zmax *= 16;
		
		int maxDimension = Math.max(Chunk.Y_MAX, Math.max(xmax-xmin, zmax-zmin));
		int requiredDepth = QuickMath.log2(QuickMath.nextPow2(maxDimension));
		
		// create new octree to fit all chunks
		octree = new Octree(requiredDepth);
		
		int xroom = (1<<requiredDepth)-(xmax-xmin);
		int yroom = (1<<requiredDepth)-Chunk.Y_MAX;
		int zroom = (1<<requiredDepth)-(zmax-zmin);
		
		origin.set(xmin - xroom/2, -yroom/2, zmin - zroom/2);
		
		if (controller != null && controller.waterHeight > 0) {
			for (int x = 0; x < (1<<octree.depth); ++x) {
				for (int z = 0; z < (1<<octree.depth); ++z) {
					for (int y = yroom/2; y < (yroom/2)+controller.waterHeight-1; ++y) {
						octree.set(Block.WATER.id | (1<<16), x, y, z);
					}
				}
			}
			for (int x = 0; x < (1<<octree.depth); ++x) {
				for (int z = 0; z < (1<<octree.depth); ++z) {
					octree.set(Block.WATER.id, x, (yroom/2)+controller.waterHeight-1, z);
				}
			}
		}
		
		// parse the regions first - force chunk lists to be populated!
		Set<ChunkPosition> regions = new HashSet<ChunkPosition>();
		for (ChunkPosition cp: chunksToLoad) {
			regions.add(cp.getRegionPosition());
		}
		
		for (ChunkPosition region: regions) {
			world.regionDiscovered(region);
			world.getRegion(region).parse();
		}

		for (ChunkPosition cp : chunksToLoad) {
			
			if (loadedChunks.contains(cp))
				continue;

			loadedChunks.add(cp);

			world.getChunk(cp).getBlockData(blocks, data, biomes);
			chunks += 1;
			
			for (int cy = 0; cy < 256; ++cy) {
				for (int cz = 0; cz < 16; ++cz) {
					int z = cz + cp.z*16 - origin.z;
					for (int cx = 0; cx < 16; ++cx) {
						int x = cx + cp.x*16 - origin.x;
						int index = Chunk.chunkIndex(cx, cy, cz);
						Block block = Block.values[0xFF & blocks[index]];
						
						if (cx > 0 && cx < 15 && cz > 0 && cz < 15 && cy > 0 && cy < 255 &&
								block != Block.STONE && block.isSolid()) {
							
							// set obscured blocks to stone
							if (Block.values[0xFF & blocks[index-1]].isSolid() &&
									Block.values[0xFF & blocks[index+1]].isSolid() &&
									Block.values[0xFF & blocks[index-Chunk.X_MAX]].isSolid() &&
									Block.values[0xFF & blocks[index+Chunk.X_MAX]].isSolid() &&
									Block.values[0xFF & blocks[index-Chunk.X_MAX*Chunk.Z_MAX]].isSolid() &&
									Block.values[0xFF & blocks[index+Chunk.X_MAX*Chunk.Z_MAX]].isSolid()) {
								octree.set(Block.STONE.id, x, cy - origin.y, z);
								continue;
							}
						}
						
						int metadata = 0xFF & data[index/2];
						metadata >>= (cx % 2) * 4;
						metadata &= 0xF;
						
						if (block == Block.STATIONARYWATER)
							block = Block.WATER;
						else if (block == Block.STATIONARYLAVA)
							block = Block.LAVA;
						
						int type = block.id;
						// store metadata
						switch (block) {
						case LILYPAD:
							type = 0;
							break;
							
						case WATER:
							if (cy < 255) {
								// is there water above?
								index = Chunk.chunkIndex(cx, cy+1, cz);
								Block above = Block.values[0xFF & blocks[index]];
								if (BlockType.water.contains(above)) {
									type = type | (1<<16);
								} else if (above == Block.LILYPAD) {
									type = type | (1<<17);
								}
								type |= metadata << 8;
							}
							break;
							
						case LAVA:
							if (cy < 255) {
								// is there lava above?
								index = Chunk.chunkIndex(cx, cy+1, cz);
								if (BlockType.lava.contains(Block.values[0xFF & blocks[index]])) {
									type = type | (1 << 16);
								}
								type |= metadata << 8;
							}
							break;
							
						case GRASS:
							if (cy < 255) {
								// is it snow covered?
								index = Chunk.chunkIndex(cx, cy+1, cz);
								int blockAbove = 0xFF & blocks[index];
								if (blockAbove == Block.SNOW.id) {
									type = type | (1 << 8);// 9th bit is the snow bit
								}
							}
							// fallthrough!
							
						case TALLGRASS:
						case LEAVES:
							// save biome ID
							int biomeId = 0xFF & biomes[Chunk.chunkXZIndex(cx, cz)];
							type |= biomeId << 24;
							type |= metadata << 8;
							break;
							
						case WOODENDOOR:
						case IRONDOOR:
							if ((metadata & 8) != 0) {
								// top
								int bottom = 0;
								if (cy > 0) {
									bottom = 0xFF & data[Chunk.chunkIndex(cx, cy-1, cz)/2];
									bottom >>= (cx % 2) * 4;
									bottom &= 7;// clear top-flag
								}
								// store bottom door info in biomeId field
								type |= metadata << 8;
								type |= bottom << 24;
							} else {
								// bottom
								int top = 0;
								if (cy < 255) {
									top = 0xFF & data[Chunk.chunkIndex(cx, cy+1, cz)/2];
									top >>= (cx % 2) * 4;
									top &= 7;// clear top-flag
								}
								// store bottom door info in biomeId field
								type |= top << 8;
								type |= metadata << 24;
							}
							break;
							
						default:
							 type |= metadata << 8;
						}
						if (BlockType.emitter.contains(block))
							emitters += 1;
						if (block == Block.AIR)
							type = 0;// air should always have type 0
						octree.set(type, cx + cp.x*16 - origin.x,
								cy - origin.y, cz + cp.z*16 - origin.z);
					}
				}
			}
		}
		
		for (ChunkPosition cp : chunksToLoad) {
			
			for (int cy = 0-origin.y; cy < Chunk.Y_MAX-origin.y; ++cy) {
				for (int cz = 0; cz < 16; ++cz) {
					int z = cz + cp.z*16 - origin.z;
					for (int cx = 0; cx < 16; ++cx) {
						int x = cx + cp.x*16 - origin.x;
						int type = octree.get(x, cy, z);
						Block block = Block.values[type & 0xFF];
						
						if ((cx == 0 || cx == 15 || cz == 0 || cz == 15) &&
								cy > -origin.y && cy < Chunk.Y_MAX-origin.y-1 &&
								block != Block.STONE && block.isSolid()) {
							if (Block.values[0xFF & octree.get(x-1, cy, z)].isSolid() &&
									Block.values[0xFF & octree.get(x+1, cy, z)].isSolid() &&
									Block.values[0xFF & octree.get(x, cy-1, z)].isSolid() &&
									Block.values[0xFF & octree.get(x, cy+1, z)].isSolid() &&
									Block.values[0xFF & octree.get(x, cy, z-1)].isSolid() &&
									Block.values[0xFF & octree.get(x, cy, z+1)].isSolid()) {
								octree.set(Block.STONE.id, x, cy, z);
								continue;
							}
						}
						
						Block other;
						switch (block) {
						case MELONSTEM:
							if (Block.values[0xFF & octree.get(x - 1, cy, z)] == Block.MELON)
								type |= 1 << 16;
							else if (Block.values[0xFF & octree.get(x + 1, cy, z)] == Block.MELON)
								type |= 2 << 16;
							else if (Block.values[0xFF & octree.get(x, cy, z - 1)] == Block.MELON)
								type |= 3 << 16;
							else if (Block.values[0xFF & octree.get(x, cy, z + 1)] == Block.MELON)
								type |= 4 << 16;
							octree.set(type, x, cy, z);
							break;
						case PUMPKINSTEM:
							if (Block.values[0xFF & octree.get(x - 1, cy, z)] == Block.PUMPKIN)
								type |= 1 << 16;
							else if (Block.values[0xFF & octree.get(x + 1, cy, z)] == Block.PUMPKIN)
								type |= 2 << 16;
							else if (Block.values[0xFF & octree.get(x, cy, z - 1)] == Block.PUMPKIN)
								type |= 3 << 16;
							else if (Block.values[0xFF & octree.get(x, cy, z + 1)] == Block.PUMPKIN)
								type |= 4 << 16;
							octree.set(type, x, cy, z);
							break;
						case CHEST:
							int dir = type >> 8;
							int tex = 0;
							if (dir < 4) {
								if (Block.values[0xFF & octree.get(x - 1, cy, z)] == Block.CHEST)
									tex = 1 + (dir-1) % 2;
								else if (Block.values[0xFF & octree.get(x + 1, cy, z)] == Block.CHEST)
									tex = 1 + dir % 2;
							} else {
								if (Block.values[0xFF & octree.get(x, cy, z - 1)] == Block.CHEST)
									tex = 1 + dir % 2;
								else if (Block.values[0xFF & octree.get(x, cy, z + 1)] == Block.CHEST)
									tex = 1 + (dir-1) % 2;
							}
							type |= tex << 16;
							octree.set(type, x, cy, z);
							break;
						case IRONBARS:
							other = Block.values[0xFF & octree.get(x, cy, z - 1)];
							if (BlockType.ironBarConnector.contains(other))
								type |= 1 << 8;
							other = Block.values[0xFF & octree.get(x, cy, z + 1)];
							if (BlockType.ironBarConnector.contains(other))
								type |= 2 << 8;
							other = Block.values[0xFF & octree.get(x + 1, cy, z)];
							if (BlockType.ironBarConnector.contains(other))
								type |= 4 << 8;
							other = Block.values[0xFF & octree.get(x - 1, cy, z)];
							if (BlockType.ironBarConnector.contains(other))
								type |= 8 << 8;
							octree.set(type, x, cy, z);
							break;
						case GLASSPANE:
							other = Block.values[0xFF & octree.get(x, cy, z - 1)];
							if (BlockType.glassPaneConnector.contains(other))
								type |= 1 << 8;
							other = Block.values[0xFF & octree.get(x, cy, z + 1)];
							if (BlockType.glassPaneConnector.contains(other))
								type |= 2 << 8;
							other = Block.values[0xFF & octree.get(x + 1, cy, z)];
							if (BlockType.glassPaneConnector.contains(other))
								type |= 4 << 8;
							other = Block.values[0xFF & octree.get(x - 1, cy, z)];
							if (BlockType.glassPaneConnector.contains(other))
								type |= 8 << 8;
							octree.set(type, x, cy, z);
							break;
						case FENCE:
							other = Block.values[0xFF & octree.get(x, cy, z - 1)];
							if (BlockType.fenceConnector.contains(other))
								type |= 1 << 8;
							other = Block.values[0xFF & octree.get(x, cy, z + 1)];
							if (BlockType.fenceConnector.contains(other))
								type |= 2 << 8;
							other = Block.values[0xFF & octree.get(x + 1, cy, z)];
							if (BlockType.fenceConnector.contains(other))
								type |= 4 << 8;
							other = Block.values[0xFF & octree.get(x - 1, cy, z)];
							if (BlockType.fenceConnector.contains(other))
								type |= 8 << 8;
							octree.set(type, x, cy, z);
							fence++;
							break;
						case NETHERBRICKFENCE:
							other = Block.values[0xFF & octree.get(x, cy, z - 1)];
							if (BlockType.netherBrickFenceConnector.contains(other))
								type |= 1 << 8;
							other = Block.values[0xFF & octree.get(x, cy, z + 1)];
							if (BlockType.netherBrickFenceConnector.contains(other))
								type |= 2 << 8;
							other = Block.values[0xFF & octree.get(x + 1, cy, z)];
							if (BlockType.netherBrickFenceConnector.contains(other))
								type |= 4 << 8;
							other = Block.values[0xFF & octree.get(x - 1, cy, z)];
							if (BlockType.netherBrickFenceConnector.contains(other))
								type |= 8 << 8;
							octree.set(type, x, cy, z);
							fence++;
							break;
						}
					}
				}
			}
		}
		
		if (controller != null)
			controller.appendStatusText(
				String.format("Loaded %d chunks, %d emitters, %d fence posts%n",
				chunks, emitters, fence));
	}

	public Collection<ChunkPosition> loadedChunks() {
		return loadedChunks;
	}

	/**
	 * Calculate a camera position centered above all loaded chunks.
	 * @return The calculated camera position
	 */
	public Vector3d calcCenterCamera() {
		if (loadedChunks.isEmpty())
			return new Vector3d(0, 128, 0);
		
		int xmin = Integer.MAX_VALUE;
		int xmax = Integer.MIN_VALUE;
		int zmin = Integer.MAX_VALUE;
		int zmax = Integer.MIN_VALUE;
		for (ChunkPosition cp: loadedChunks) {
			if (cp.x < xmin)
				xmin = cp.x;
			if (cp.x > xmax)
				xmax = cp.x;
			if (cp.z < zmin)
				zmin = cp.z;
			if (cp.z > zmax)
				zmax = cp.z;
		}
		xmax += 1;
		zmax += 1;
		xmin *= 16;
		xmax *= 16;
		zmin *= 16;
		zmax *= 16;
		return new Vector3d((xmax + xmin)/2, 128, (zmax + zmin)/2);
	}

	public void setBiomeColorsEnabled(boolean value) {
		if (value != biomeColors) {
			biomeColors = value;
			refresh();
		}
	}
	
	public boolean getBiomeColorsEnabled() {
		return Biomes.biomeColorsEnabled;
	}

	public void centerCamera() {
		camera.setPosition(calcCenterCamera());
	}

	public Camera camera() {
		return camera;
	}

	public String name() {
		return name;
	}
	
	public boolean pathTrace() {
		return pathTrace ;
	}

	public Sky sky() {
		return sky;
	}

	public Sun sun() {
		return sun;
	}
	
	/**
	 * Toggle Monte Carlo path tracing.
	 */
	public synchronized void toggleMonteCarlo() {
		pathTrace = !pathTrace;
		refresh();
	}
	
	synchronized void waitOnRefreshRequest() throws InterruptedException {
		while ((!pathTrace || pauseRender) && !refresh)
			wait();
		refresh = false;
	}
	
	synchronized boolean shouldRefresh() {
		boolean value = refresh;
		refresh = false;
		return value;
	}
	
	public synchronized void refresh() {
		refresh = true;
		pauseRender = false;
		notifyAll();
	}

	public synchronized void pauseWait() throws InterruptedException {
		synchronized (this) {
			while (pauseRender)
				wait();
		}
	}
	
	public synchronized boolean isPaused() {
		return pauseRender;
	}

	/**
 	 * Start rendering the scene.
 	 */
	public synchronized void startRender() {
		if (!pathTrace) {
			pathTrace = true;
			pauseRender = false;
			refresh();
		}
	}
	
	/**
 	 * Pause the renderer.
 	 */
	public synchronized void pauseRender() {
		pauseRender = true;
	}
	
	/**
 	 * Resume a paused render.
 	 */
	public synchronized void resumeRender() {
		pauseRender = false;
		notifyAll();
	}

	/**
 	 * Halt the rendering process.
 	 * Puts the renderer back in preview mode.
 	 */
	public synchronized void haltRender() {
		if (pathTrace) {
			pathTrace = false;
			pauseRender = false;
			refresh();
		}
	}
	
	public World world() {
		return world;
	}

	public void moveCameraToPlayer() {
		camera.moveToPlayer(world);
	}

	public boolean stillWaterEnabled() {
		return stillWater;
	}
	
	public boolean biomeColorsEnabled() {
		return biomeColors;
	}

	public synchronized void setRayDepth(int value) {
		if (rayDepth != value) {
			rayDepth = value;
			refresh();
		}
	}

	public int getRayDepth() {
		return rayDepth;
	}

	public void setRefreshed() {
		refresh = false;
	}

	public void trace(Ray ray) {
		ray.d.set(0, -1, 0);
		ray.x.set(camera.getPosition());
		ray.x.x -= origin.x;
		ray.x.y -= origin.y;
		ray.x.z -= origin.z;
		camera.transform(ray.d);
		while (intersect(ray)) {
			if (ray.currentMaterial != 0) {
				ray.hit = true;
				break;
			}
		}
	}

	public void autoFocus() {
		Ray ray = RayPool.getDefaultRayPool().get();
		trace(ray);
		if (!ray.hit) {
			camera.setInfDof(true);
		} else {
			camera.setFocalOffset(ray.distance);
			camera.setDof(ray.distance);
			camera.setInfDof(false);
		}
		RayPool.getDefaultRayPool().dispose(ray);
	}

	public Octree getOctree() {
		return octree;
	}

	public Vector3i getOrigin() {
		return origin;
	}

	public void setClearWater(boolean value) {
		if (value != clearWater) {
			clearWater  = value;
			refresh();
		}
	}
	
	public boolean getClearWater() {
		return clearWater;
	}

	public void setName(String text) {
		if (text.length() > 0)
			name = text;
	}
	
	public boolean getGammaCorrectionEnabled() {
		return gammaCorrection;
	}

	public synchronized void setGammaCorrectionEnabled(boolean value) {
		gammaCorrection  = value;
		if (!pathTrace)
			refresh();
	}

	public double getEmitterIntensity() {
		return emitterIntensity;
	}
	
	public void setEmitterIntensity(double value) {
		emitterIntensity  = value;
		refresh();
	}

	public void setAtmosphereEnabled(boolean value) {
		if (value != atmosphereEnabled) {
			atmosphereEnabled = value;
			refresh();
		}
	}
	
	public void setVolumetricFogEnabled(boolean value) {
		if (value != volumetricFogEnabled) {
			volumetricFogEnabled = value;
			refresh();
		}
	}

	public boolean atmosphereEnabled() {
		return atmosphereEnabled;
	}

	public boolean volumetricFogEnabled() {
		return volumetricFogEnabled;
	}
}
