/* Copyright (c) 2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.chunky.renderer.cl;

import static org.jocl.CL.*;

import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferInt;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.Random;
import java.util.logging.Level;

import org.jocl.Pointer;
import org.jocl.Sizeof;
import org.jocl.cl_command_queue;
import org.jocl.cl_context;
import org.jocl.cl_context_properties;
import org.jocl.cl_device_id;
import org.jocl.cl_kernel;
import org.jocl.cl_mem;
import org.jocl.cl_platform_id;
import org.jocl.cl_program;

import se.llbit.chunky.main.Chunky;
import se.llbit.chunky.renderer.Scene;
import se.llbit.chunky.renderer.ui.Canvas3D;
import se.llbit.chunky.world.ChunkPosition;
import se.llbit.chunky.world.World;

import se.llbit.j99.castor.Symbol;
import se.llbit.j99.fragment.CommentFragmenter;
import se.llbit.j99.fragment.CompositeFragment;
import se.llbit.j99.fragment.FragmentReader;
import se.llbit.j99.fragment.IFragment;
import se.llbit.j99.fragment.IFragmenter;
import se.llbit.j99.fragment.LineFragment;
import se.llbit.j99.fragment.LineFragmenter;
import se.llbit.j99.fragment.LineSplicer;
import se.llbit.j99.fragment.TrigraphReplacer;
import se.llbit.j99.pp.Identifier;
import se.llbit.j99.pp.Macro;
import se.llbit.j99.pp.ObjMacro;
import se.llbit.j99.pp.PPState;
import se.llbit.j99.pp.PPVisitor;
import se.llbit.j99.pp.SourceFile;
import se.llbit.j99.pp.StringLit;
import se.llbit.j99.problem.CompileProblem;
import se.llbit.j99.util.DirectiveParser;
import se.llbit.math.Color;
import se.llbit.math.Matrix3d;
import se.llbit.math.Octree;
import se.llbit.math.Vector3d;

public class RenderManager extends Thread {
	
	private static final int WORK_X = 50;
	private static final int WORK_Y = 50;
	private static final int WORK_SIZE = WORK_X*WORK_Y;
	private cl_kernel kernel;
	private cl_mem sampleBuffer;
	private cl_mem octreeBuffer;
	private cl_mem transformBuffer;
	private cl_mem originBuffer;
	private cl_mem seedBuffer;
	private cl_command_queue commandQueue;
	private int bufferWidth;
	private int bufferHeight;
	private BufferedImage buffer;
	private Canvas3D canvas;
	private Octree octree;
	private final Vector3d origin = new Vector3d();
	private final Vector3d up = new Vector3d(0, 1, 0);
	private double pitch = 0;
	private double yaw = - Math.PI / 2;
	private final Matrix3d transform = new Matrix3d();
	private final Matrix3d tmpTransform = new Matrix3d();
	private int numSamples = 0;
	private DataBufferInt dataBuffer;
	private int[] imgData;
	
	public RenderManager(Canvas3D canvas, String[] args) throws InterruptedException {
		super("Render Manager");
		
		this.canvas = canvas;
		bufferWidth = canvas.getWidth();
		bufferHeight = canvas.getHeight();
		buffer = new BufferedImage(bufferWidth, bufferHeight,
				BufferedImage.TYPE_INT_ARGB);
		dataBuffer = (DataBufferInt) buffer.getRaster().getDataBuffer();
        imgData = dataBuffer.getData();
		setupOpenCL(args);
	}
	
	@SuppressWarnings("unused")
	private void setupOpenCL(String[] args) throws InterruptedException {
		// don't want to do error checks
		setExceptionsEnabled(true);
		
		int[] num_platforms = new int[1];
		clGetPlatformIDs(0, null, num_platforms);
		
		if (num_platforms[0] == 0) {
			Chunky.logger.severe("No OpenCL platforms available!");
			return;
		}
		
		cl_platform_id[] platforms = new cl_platform_id[num_platforms[0]];
		clGetPlatformIDs(platforms.length, platforms, null);
		
		for (int i = 0; i < platforms.length; ++i) {
			byte[] data = new byte[500];
			Pointer dataPtr = Pointer.to(data);
			long[] dataSize = new long[1];
			
			clGetPlatformInfo(platforms[i], CL_PLATFORM_NAME, 500, dataPtr, dataSize);
			String name = new String(data, 0, (int) dataSize[0] - 1);
			
			clGetPlatformInfo(platforms[i], CL_PLATFORM_VERSION, 500, dataPtr, dataSize);
			String version = new String(data, 0, (int) dataSize[0] - 1);
			
			System.out.println(String.format("Platform %d: %s (%s)", i, name, version));
		}
		
		int[] num_devices = new int[1];
		clGetDeviceIDs(platforms[0], CL_DEVICE_TYPE_ALL, 0, null, num_devices);
		
		if (num_devices[0] == 0) {
			Chunky.logger.severe("No devices available for the platform!");
			return;
		}
		
		cl_device_id[] devices = new cl_device_id[num_devices[0]];
		clGetDeviceIDs(platforms[0], CL_DEVICE_TYPE_ALL, num_devices[0], devices, null);
		
		for (int i = 0; i < devices.length; ++i) {
			long[] cl_long = new long[1];
			int[] cl_uint = new int[1];
			byte[] data = new byte[500];
			Pointer dataPtr = Pointer.to(data);
			long[] dataSize = new long[1];
			clGetDeviceInfo(devices[i], CL_DEVICE_TYPE, Sizeof.cl_long, Pointer.to(cl_long), null);
			
			String deviceType = "Unknown";
			switch ((int) cl_long[0]) {
			case (int) CL_DEVICE_TYPE_CPU:
				deviceType = "CPU";
				break;
			case (int) CL_DEVICE_TYPE_GPU:
				deviceType = "GPU";
				break;
			case (int) CL_DEVICE_TYPE_ACCELERATOR:
				deviceType = "ACCELERATOR";
				break;
			case (int) CL_DEVICE_TYPE_DEFAULT:
				deviceType = "DEFAULT";
				break;
			}
			
			clGetDeviceInfo(devices[i], CL_DEVICE_NAME, 500, dataPtr, dataSize);
			String name = new String(data, 0, (int) dataSize[0] - 1);
			
			clGetDeviceInfo(devices[i], CL_DEVICE_MAX_COMPUTE_UNITS, Sizeof.cl_uint, Pointer.to(cl_uint), null);
			
			int computeUnits = cl_uint[0];
			
			clGetDeviceInfo(devices[i], CL_DEVICE_MAX_WORK_GROUP_SIZE, Sizeof.size_t, Pointer.to(cl_uint), null);
			
			int workGroupSize = cl_uint[0];

			System.out.println(String.format("Device %d: Type = %s", i, deviceType));
			System.out.println(String.format("Device %d: Name = %s", i, name));
			System.out.println(String.format("Device %d: Max compute units = %d", i, computeUnits));
			System.out.println(String.format("Device %d: Max work group size = %d", i, workGroupSize));
			
			cl_context_properties contextProps = new cl_context_properties();
			contextProps.addProperty(CL_CONTEXT_PLATFORM, platforms[0]);
			
			cl_context context = clCreateContext(contextProps, 1, new cl_device_id[] {devices[0]},
					null, null, null);
			commandQueue = clCreateCommandQueue(context, devices[0], 0, null);
			
			String kernelSource = processKernel();
			cl_program program = clCreateProgramWithSource(context, 1,
					new String[] {kernelSource}, null, null);
			clBuildProgram(program, 0, null, "-cl-mad-enable", null, null);
			kernel = clCreateKernel(program, "path_trace", null);
			sampleBuffer = clCreateBuffer(context, CL_MEM_READ_WRITE,
					3 * bufferWidth * bufferHeight * Sizeof.cl_float, null, null);
			
			World world = new World(new File(args[0]));
			Scene scene = new Scene(null, world);
			Collection<ChunkPosition> chunksToLoad = new LinkedList<ChunkPosition>();
			for (int x = 0; x < 10; ++x) {
				for (int y = 0; y < 10; ++y) {
					chunksToLoad.add(ChunkPosition.get(x, y));
				}
			}
			world.setDimension(0);
			scene.loadChunks(chunksToLoad);
			octree = scene.getOctree();
			origin.set(scene.getOrigin());
			origin.scale(-1);
			origin.y = 90;
			origin.x += 2;
			origin.z += 8;
	        int[] octreeData = octree.toDataBuffer();
			octreeBuffer = clCreateBuffer(context, CL_MEM_READ_ONLY,
					octreeData.length * Sizeof.cl_uint, null, null);
			clEnqueueWriteBuffer(commandQueue, octreeBuffer, CL_TRUE, 0,
					octreeData.length * Sizeof.cl_int, Pointer.to(octreeData),
	        		0, null, null);
			transformBuffer = clCreateBuffer(context, CL_MEM_READ_ONLY,
					9 * Sizeof.cl_float, null, null);
			originBuffer = clCreateBuffer(context, CL_MEM_READ_ONLY,
					3 * Sizeof.cl_float, null, null);
			seedBuffer = clCreateBuffer(context, CL_MEM_READ_WRITE,
					WORK_SIZE * Sizeof.cl_uint2, null, null);
			
			Random random = new Random(System.currentTimeMillis());
			// seed the RNGs
	        int[] buf = new int[WORK_SIZE*2];
	        for (int y = 0; y < WORK_Y; ++y) {
	        	for (int x = 0; x < WORK_X; ++x) {
	        		buf[(x + y*WORK_X)*2] = random.nextInt();
	        		buf[(x + y*WORK_X)*2 + 1] = random.nextInt();
	        	}
	        }
	        
	        clEnqueueWriteBuffer(commandQueue, seedBuffer, CL_TRUE, 0,
	        		WORK_SIZE * Sizeof.cl_uint2, Pointer.to(buf),
	        		0, null, null);
			updateTransform();
			updateOrigin();
			break;
		}
	}

	private String processKernel() {
		//Preprocessor preprocessor = new Preprocessor(in, problems, includeDirs, basePath)

		try {

			String sourceFn = "pathtrace.c";
			String basePath = "";
			InputStream in = new FileInputStream(sourceFn);
			ByteArrayOutputStream out = new ByteArrayOutputStream();

			ArrayList<CompileProblem> problems = new ArrayList<CompileProblem>();
			IFragmenter fragmenter = new CommentFragmenter(new LineSplicer(new TrigraphReplacer(
							new LineFragmenter(sourceFn, in))), problems);

			PPState state = new PPState();
			state.setProblemCollection(problems);
			state.setIncludeDirs(new LinkedList<String>());
			// predefined macros
			// TODO: these need to be attached to the root tree
			state.define(new Identifier("_WIN32_"),
					new ObjMacro(new Identifier("_WIN32_"),
					new se.llbit.j99.pp.List<se.llbit.j99.pp.Token>()));

			IFragment processed = preprocess(fragmenter, state, sourceFn, basePath);
			
			in.close();

			CompileProblem.reportProblems(problems);
			if (!CompileProblem.isCritical(problems)) {
				
				Reader reader = new FragmentReader(processed);

				while (reader.ready())
					out.write((char)reader.read());
				out.close();
				String kernel = new String(out.toByteArray());
				return kernel;
			}

		} catch (IOException e) {
			Chunky.logger.log(Level.SEVERE, "Fatal IO error", e);
		} catch (RuntimeException e) {
			Chunky.logger.log(Level.SEVERE, "Fatal error", e);
		}

		return null;
	}
	
	private static IFragment preprocess(IFragmenter in, PPState state, String fn, String basePath) {
		Identifier name = new Identifier("__FILE__");
		StringLit filename = new StringLit("\""+fn+"\"");
		name.setToken(new Symbol(new LineFragment("@j99", 0, 0, "__FILE__")));
		filename.setToken(new Symbol(new LineFragment("@j99", 0, 0, "\""+fn+"\"")));
		Macro fileMacro = new ObjMacro(name,
				new se.llbit.j99.pp.List<se.llbit.j99.pp.Token>().add(filename));// TODO: this needs to be attached to the root tree
		state.define(name, fileMacro);

		DirectiveParser parser = new DirectiveParser(in, state);
		SourceFile root = parser.parse();
		PPVisitor visitor = new PPVisitor(state, basePath);
		root.accept(visitor);
		return new CompositeFragment(visitor.getResult());
	}
	
	public void run() {

		while (!isInterrupted()) {
			long globalWorkSize[] = new long[2];
	        globalWorkSize[0] = WORK_X;
	        globalWorkSize[1] = WORK_Y;
	        
	        synchronized (this) {
	        
		        clSetKernelArg(kernel, 0, Sizeof.cl_mem, Pointer.to(sampleBuffer));
		        clSetKernelArg(kernel, 1, Sizeof.cl_uint, Pointer.to(new int[] {bufferWidth}));
		        clSetKernelArg(kernel, 2, Sizeof.cl_uint, Pointer.to(new int[] {bufferHeight}));
		        clSetKernelArg(kernel, 3, Sizeof.cl_uint, Pointer.to(new int[] {WORK_X}));
		        clSetKernelArg(kernel, 4, Sizeof.cl_uint, Pointer.to(new int[] {WORK_Y}));
		        clSetKernelArg(kernel, 5, Sizeof.cl_mem, Pointer.to(originBuffer));
		        clSetKernelArg(kernel, 6, Sizeof.cl_mem, Pointer.to(octreeBuffer));
		        clSetKernelArg(kernel, 7, Sizeof.cl_uint, Pointer.to(new int[] {octree.depth}));
		        clSetKernelArg(kernel, 8, Sizeof.cl_mem, Pointer.to(transformBuffer));
		        clSetKernelArg(kernel, 9, Sizeof.cl_mem, Pointer.to(seedBuffer));
		        clSetKernelArg(kernel, 10, Sizeof.cl_uint, Pointer.to(new int[] {numSamples}));
	
		        clEnqueueNDRangeKernel(commandQueue, kernel, 2, null, 
		                globalWorkSize, null, 0, null, null);
		        
		        numSamples += 1;
	        }
	        
	        updateCanvas();
		}
	}
	
	public synchronized void refresh() {
		numSamples = 0;
	}
	
	public synchronized void updateTransform() {
		tmpTransform.rotZ(pitch);
		transform.rotY(yaw);
		transform.mul(tmpTransform);
		float[] mat = new float[9];
		mat[0] = (float) transform.m11;
		mat[1] = (float) transform.m12;
		mat[2] = (float) transform.m13;
		mat[3] = (float) transform.m21;
		mat[4] = (float) transform.m22;
		mat[5] = (float) transform.m23;
		mat[6] = (float) transform.m31;
		mat[7] = (float) transform.m32;
		mat[8] = (float) transform.m33;
		clEnqueueWriteBuffer(commandQueue, transformBuffer, CL_TRUE, 0,
				mat.length * Sizeof.cl_float, Pointer.to(mat),
        		0, null, null);
	}
	
	public synchronized void updateOrigin() {
		float[] o = new float[3];
		o[0] = (float) origin.x;
		o[1] = (float) origin.y;
		o[2] = (float) origin.z;
		clEnqueueWriteBuffer(commandQueue, originBuffer, CL_TRUE, 0,
				o.length * Sizeof.cl_float, Pointer.to(o),
        		0, null, null);
	}

	private void updateCanvas() {
		if (canvas.isDisplayable()) {
			try {
		        float[] samples = new float[bufferWidth*bufferHeight*3];
		        clEnqueueReadBuffer(commandQueue, sampleBuffer, CL_TRUE, 0,
		        		3 * bufferWidth * bufferHeight * Sizeof.cl_float, Pointer.to(samples),
		        		0, null, null);
		        for (int i = 0; i < bufferWidth*bufferHeight; ++i) {
		        	imgData[i] = Color.getRGB(
		        			Math.min(1, samples[i*3]),
		        			Math.min(1, samples[i*3+1]),
        					Math.min(1, samples[i*3+2]));
		        }
		        
				Graphics g = canvas.getGraphics();
				g.drawImage(buffer,
						0, 0, canvas.getWidth(), canvas.getHeight(),
						0, 0, bufferWidth, bufferHeight, null);
				g.dispose();
				canvas.update();
			} catch (IllegalStateException e) {
				
			}
		}
	}

	public synchronized void moveForward(double v) {
		Vector3d d = new Vector3d(0, -1, 0);
		transform.transform(d);
		origin.scaleAdd(v, d, origin);
		updateOrigin();
		refresh();
	}
	
	public synchronized void moveBackward(double v) {
		Vector3d d = new Vector3d(0, -1, 0);
		transform.transform(d);
		origin.scaleAdd(-v, d, origin);
		updateOrigin();
		refresh();
	}
	
	public synchronized void moveUp(double v) {
		origin.scaleAdd(v, up, origin);
		updateOrigin();
		refresh();
	}
	
	public synchronized void moveDown(double v) {
		origin.scaleAdd(-v, up, origin);
		updateOrigin();
		refresh();
	}
	
	public synchronized void strafeLeft(double v) {
		Vector3d d = new Vector3d(1, 0, 0);
		tmpTransform.rotY(yaw);
		tmpTransform.transform(d);
		Vector3d right = new Vector3d();
		right.cross(up, d);
		origin.scaleAdd(-v, right, origin);
		updateOrigin();
		refresh();
	}
	
	public synchronized void strafeRight(double v) {
		Vector3d d = new Vector3d(1, 0, 0);
		tmpTransform.rotY(yaw);
		tmpTransform.transform(d);
		Vector3d right = new Vector3d();
		right.cross(up, d);
		origin.scaleAdd(v, right, origin);
		updateOrigin();
		refresh();
	}
	
	public synchronized void rotateView(double yaw, double pitch) {
		double fov = 70;
		double fovRad = (fov / 360) * Math.PI;
		this.yaw += yaw * fovRad;
		this.pitch += pitch * fovRad;
		
		this.pitch = Math.min(0, this.pitch);
		this.pitch = Math.max(-Math.PI, this.pitch);
		
		if (this.yaw > Math.PI * 2)
			this.yaw -= Math.PI * 2;
		else if (this.yaw < -Math.PI * 2)
			this.yaw += Math.PI * 2;
		
		updateTransform();
		refresh();
	}
}
