/* Copyright (c) 2012 Jesper Öqvist <jesper@llbit.se>
 *
 * This file is part of Chunky.
 *
 * Chunky is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Chunky is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Chunky.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.llbit.util;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Properties;

/**
 * Utility class for managing program properties.
 * 
 * @author Jesper Öqvist <jesper@llbit.se>
 */
public final class ProgramProperties {
	
	private static Properties properties = new Properties();
	private static boolean modified = false;
	
	private ProgramProperties() { }
	
	static {
		loadProperties();
	}

	/**
	 * Load saved properties.
	 */
	private static void loadProperties() {
		try {
			InputStream in = new FileInputStream("settings.cfg");
			properties.load(in);
		} catch (IOException e) {
			// no saved settings available - defaults will be used
		}
	}
	
	/**
	 * Save settings to file.
	 * @param comment Comment
	 */
	public static void saveProperties(String comment) {
		if (modified) {
			try {
				OutputStream out = new FileOutputStream("settings.cfg");
				properties.store(out, comment);
			} catch (IOException e1) {
				e1.printStackTrace();
			}
		}
	}
	
	/**
	 * Get a property value
	 * @param name Property name
	 * @param defValue Default value
	 * @return The value of the property, or <code>defValue</code> if
	 * the property was not set.
	 */
	public static String getProperty(String name, String defValue) {
		return properties.getProperty(name, defValue);
	}
	
	/**
	 * Get the integer value of a property
	 * @param name Property name
	 * @param defValue Default value
	 * @return The value of the property, or <code>defValue</code> if
	 * the property was not set, or was not set to a valid integer value.
	 */
	public static int getIntProperty(String name, int defValue) {
		try {
			return Integer.parseInt(properties.getProperty(name, ""+defValue));
		} catch (NumberFormatException e) {
			return defValue;
		}
	}
	
	/**
	 * Set the value of a property
	 * @param name Property name
	 * @param value Property value
	 */
	public static void setProperty(String name, String value) {
		// only set the property if the value has changed
		if ((value == null && properties.getProperty(name) != null)
				|| (!value.equals(properties.getProperty(name)))) {
			
			modified = true;
			properties.setProperty(name, value);
		}
	}

	/**
	 * Get a propertie's value
	 * @param name Property name
	 * @return The propertie's value, or <code>null</code> if the property was not set.
	 */
	public static String getProperty(String name) {
		return properties.getProperty(name);
	}

	/**
	 * Clears the property
	 * @param name
	 */
	public static void removeProperty(String name) {
		if (properties.containsKey(name)) {
			properties.remove(name);
			modified = true;
		}
	}
}
